import atexit
import contextlib
import logging
import os
import pathlib
import pprint
import re
import shutil
import tarfile
import textwrap
import time
from typing import TYPE_CHECKING, Any, Dict, List
from zipfile import ZipFile

import attr
import distro
import packaging
import psutil
import pytest
import requests
from pytestshellutils.shell import DaemonImpl, Subprocess
from pytestshellutils.utils.processes import (
    ProcessResult,
    _get_cmdline,
    terminate_process,
)
from pytestskipmarkers.utils import platform
from saltfactories.bases import SystemdSaltDaemonImpl
from saltfactories.cli import call, key, salt
from saltfactories.daemons import api, master, minion
from saltfactories.utils import cli_scripts

try:
    import crypt

    HAS_CRYPT = True
except ImportError:
    HAS_CRYPT = False
try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False

try:
    import winreg

    HAS_WINREG = True
except ImportError:
    HAS_WINREG = False

TESTS_DIR = pathlib.Path(__file__).resolve().parent.parent
CODE_DIR = TESTS_DIR.parent
ARTIFACTS_DIR = CODE_DIR / "artifacts"

log = logging.getLogger(__name__)


@attr.s(kw_only=True, slots=True)
class SaltPkgInstall:
    conf_dir: pathlib.Path = attr.ib()
    system_service: bool = attr.ib(default=False)
    proc: Subprocess = attr.ib(init=False, repr=False)
    pkgs: List[str] = attr.ib(factory=list)
    onedir: bool = attr.ib(default=False)
    singlebin: bool = attr.ib(default=False)
    compressed: bool = attr.ib(default=False)
    hashes: Dict[str, Dict[str, Any]] = attr.ib(repr=False)
    root: pathlib.Path = attr.ib(default=None)
    run_root: pathlib.Path = attr.ib(default=None)
    ssm_bin: pathlib.Path = attr.ib(default=None)
    bin_dir: pathlib.Path = attr.ib(default=None)
    # The artifact is an installer (exe, msi, pkg, rpm, deb)
    installer_pkg: bool = attr.ib(default=False)
    upgrade: bool = attr.ib(default=False)
    # install salt or not. This allows someone
    # to test a currently installed version of salt
    no_install: bool = attr.ib(default=False)
    no_uninstall: bool = attr.ib(default=False)

    distro_id: str = attr.ib(init=False)
    distro_codename: str = attr.ib(init=False)
    distro_name: str = attr.ib(init=False)
    distro_version: str = attr.ib(init=False)
    pkg_mngr: str = attr.ib(init=False)
    rm_pkg: str = attr.ib(init=False)
    salt_pkgs: List[str] = attr.ib(init=False)
    install_dir: pathlib.Path = attr.ib(init=False)
    binary_paths: Dict[str, List[pathlib.Path]] = attr.ib(init=False)
    classic: bool = attr.ib(default=False)
    prev_version: str = attr.ib()
    pkg_version: str = attr.ib(default="1")
    repo_data: str = attr.ib(init=False, repr=False)
    major: str = attr.ib(init=False)
    minor: str = attr.ib(init=False)
    relenv: bool = attr.ib(default=True)
    file_ext: bool = attr.ib(default=None)
    config_path: str = attr.ib(init=False)

    @proc.default
    def _default_proc(self):
        return Subprocess()

    @hashes.default
    def _default_hashes(self):
        return {
            "BLAKE2B": {"file": None, "tool": "-blake2b512"},
            "SHA3_512": {"file": None, "tool": "-sha3-512"},
            "SHA512": {"file": None, "tool": "-sha512"},
        }

    @distro_id.default
    def _default_distro_id(self):
        return distro.id().lower()

    @distro_codename.default
    def _default_distro_codename(self):
        return distro.codename().lower()

    @distro_name.default
    def _default_distro_name(self):
        if distro.name():
            return distro.name().split()[0].lower()

    @distro_version.default
    def _default_distro_version(self):
        return distro.version().lower()

    @pkg_mngr.default
    def _default_pkg_mngr(self):
        if self.distro_id in ("centos", "redhat", "amzn", "fedora", "photon"):
            return "yum"
        elif self.distro_id in ("ubuntu", "debian"):
            ret = self.proc.run("apt-get", "update")
            self._check_retcode(ret)
            return "apt-get"

    @rm_pkg.default
    def _default_rm_pkg(self):
        if self.distro_id in ("centos", "redhat", "amzn", "fedora", "photon"):
            return "remove"
        elif self.distro_id in ("ubuntu", "debian"):
            return "purge"

    @salt_pkgs.default
    def _default_salt_pkgs(self):
        salt_pkgs = [
            "salt-api",
            "salt-syndic",
            "salt-ssh",
            "salt-master",
            "salt-cloud",
            "salt-minion",
        ]
        if self.distro_id in ("centos", "redhat", "amzn", "fedora", "photon"):
            salt_pkgs.append("salt")
        elif self.distro_id in ("ubuntu", "debian"):
            salt_pkgs.append("salt-common")
        return salt_pkgs

    @install_dir.default
    def _default_install_dir(self):
        if platform.is_windows():
            install_dir = pathlib.Path(
                os.getenv("ProgramFiles"), "Salt Project", "Salt"
            ).resolve()
        elif platform.is_darwin():
            # TODO: Add mac install dir path
            install_dir = pathlib.Path("/opt", "salt")
        else:
            install_dir = pathlib.Path("/opt", "saltstack", "salt")
        return install_dir

    @config_path.default
    def _default_config_path(self):
        """
        Default location for salt configurations
        """
        if platform.is_windows():
            config_path = pathlib.Path("C://salt", "etc", "salt")
        else:
            config_path = pathlib.Path("/etc", "salt")
        return config_path

    @repo_data.default
    def _default_repo_data(self):
        """
        Query to see the published Salt artifacts
        from repo.json
        """
        url = "https://repo.saltproject.io/salt/onedir/repo.json"
        ret = requests.get(url)
        data = ret.json()
        return data

    def check_relenv(self, version):
        """
        Detects if we are using relenv
        onedir build
        """
        relenv = False
        if packaging.version.parse(version) >= packaging.version.parse("3006.0"):
            relenv = True
        return relenv

    def update_process_path(self):
        # The installer updates the path for the system, but that doesn't
        # make it to this python session, so we need to update that
        os.environ["PATH"] = ";".join([str(self.install_dir), os.getenv("path")])

        # When the MSI installer is run from self.proc.run, it doesn't update
        # the registry. When run from a normal command prompt it does. Until we
        # figure that out, we will update the process path as above. This
        # doesn't really check that the path is being set though... but I see
        # no other way around this
        # if HAS_WINREG:
        #     log.debug("Refreshing the path")
        #     # Get the updated system path from the registry
        #     path_key = winreg.OpenKeyEx(
        #         winreg.HKEY_LOCAL_MACHINE,
        #         r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment",
        #     )
        #     current_path = winreg.QueryValueEx(path_key, "path")[0]
        #     path_key.Close()
        #     # Update the path for the current running process
        #     os.environ["PATH"] = current_path

    def get_version(self, version_only=False):
        """
        Return the version information
        needed to install a previous version
        of Salt.
        """
        prev_version = self.prev_version
        pkg_version = None
        if not self.upgrade:
            # working with local artifact
            version = ""
            for artifact in ARTIFACTS_DIR.glob("**/*.*"):
                version = re.search(
                    r"([0-9].*)(\-[0-9].fc|\-[0-9].el|\+ds|\_all|\_any|\_amd64|\_arm64|\-[0-9].am|(\-[0-9]-[a-z]*-[a-z]*[0-9_]*.|\-[0-9]*.*)(tar.gz|tar.xz|zip|exe|msi|pkg|rpm|deb))",
                    artifact.name,
                )
                if version:
                    version = version.groups()[0].replace("_", "-").replace("~", "")
                    version = version.split("-")[0]
                    # TODO: Remove this clause.  This is to handle a versioning difficulty between pre-3006
                    # dev versions and older salt versions on deb-based distros
                    if version.startswith("1:"):
                        version = version[2:]
                    break
            major, minor = version.split(".", 1)
        else:
            if not prev_version:
                # We did not pass in a version, lets detect the latest
                # version information of a Salt artifact.
                latest = list(self.repo_data["latest"].keys())[0]
                version = self.repo_data["latest"][latest]["version"]
                if "-" in version:
                    prev_version, pkg_version = version.split("-")
                else:
                    prev_version, pkg_version = version, None
            else:
                # We passed in a version, but lets check if the pkg_version
                # is defined. Relenv pkgs do not define a pkg build number
                if "-" not in prev_version and not self.check_relenv(
                    version=prev_version
                ):
                    pkg_numbers = [
                        x for x in self.repo_data.keys() if prev_version in x
                    ]
                    pkg_version = 1
                    for number in pkg_numbers:
                        if "-" in number:
                            number = int(number.split("-")[1])
                            if number > pkg_version:
                                pkg_version = number
            major, minor = prev_version.split(".")
        if version_only:
            return version
        return major, minor, prev_version, pkg_version

    def __attrs_post_init__(self):
        self.major, self.minor, self.prev_version, self.pkg_version = self.get_version()
        self.relenv = self.check_relenv(self.major)
        file_ext_re = r"tar\.gz"
        if platform.is_darwin():
            file_ext_re = r"tar\.gz|pkg"
        if platform.is_windows():
            file_ext_re = "zip|exe|msi"
        for f_path in ARTIFACTS_DIR.glob("**/*.*"):
            f_path = str(f_path)
            if re.search(f"salt-(.*).({file_ext_re})$", f_path, re.IGNORECASE):
                # Compressed can be zip, tar.gz, exe, or pkg. All others are
                # deb and rpm
                self.compressed = True
                self.file_ext = os.path.splitext(f_path)[1].strip(".")
                if self.file_ext == "gz":
                    if f_path.endswith("tar.gz"):
                        self.file_ext = "tar.gz"
                self.pkgs.append(f_path)
                if platform.is_windows():
                    self.root = pathlib.Path(os.getenv("LocalAppData")).resolve()
                    if self.file_ext == "zip":
                        with ZipFile(f_path, "r") as zip:
                            first = zip.infolist()[0]
                            if first.filename == "salt/ssm.exe":
                                self.onedir = True
                                self.bin_dir = self.root / "salt" / "salt"
                                self.run_root = self.bin_dir / "salt.exe"
                                self.ssm_bin = self.root / "salt" / "ssm.exe"
                            elif first.filename == "salt.exe":
                                self.singlebin = True
                                self.run_root = self.root / "salt.exe"
                                self.ssm_bin = self.root / "ssm.exe"
                            else:
                                log.error(
                                    "Unexpected archive layout. First: %s",
                                    first.filename,
                                )
                    elif self.file_ext in ["exe", "msi"]:
                        self.compressed = False
                        self.onedir = True
                        self.installer_pkg = True
                        self.root = self.install_dir.parent
                        self.bin_dir = self.install_dir
                        self.ssm_bin = self.install_dir / "ssm.exe"
                    else:
                        log.error("Unexpected file extension: %s", self.file_ext)
                else:
                    if platform.is_darwin():
                        self.root = pathlib.Path(os.sep, "opt")
                    else:
                        self.root = pathlib.Path(os.sep, "usr", "local", "bin")

                    if self.file_ext == "pkg":
                        self.compressed = False
                        self.onedir = True
                        self.installer_pkg = True
                        self.bin_dir = self.root / "salt" / "bin"
                        self.run_root = self.bin_dir / "run"
                    elif self.file_ext == "tar.gz":
                        with tarfile.open(f_path) as tar:
                            # The first item will be called salt
                            first = next(iter(tar.getmembers()))
                            if first.name == "salt" and first.isdir():
                                self.onedir = True
                                self.bin_dir = self.root / "salt" / "run"
                                self.run_root = self.bin_dir / "run"
                            elif first.name == "salt" and first.isfile():
                                self.singlebin = True
                                self.run_root = self.root / "salt"
                            else:
                                log.error(
                                    "Unexpected archive layout. First: %s (isdir: %s, isfile: %s)",
                                    first.name,
                                    first.isdir(),
                                    first.isfile(),
                                )
                    else:
                        log.error("Unexpected file extension: %s", self.file_ext)

            if re.search(
                r"salt(.*)(x86_64|all|amd64|aarch64|arm64)\.(rpm|deb)$", f_path
            ):
                self.installer_pkg = True
                self.pkgs.append(f_path)

        if not self.pkgs:
            pytest.fail("Could not find Salt Artifacts")

        python_bin = self.install_dir / "bin" / "python3"
        if platform.is_windows():
            python_bin = self.install_dir / "Scripts" / "python.exe"
        if not self.compressed:
            if platform.is_windows():
                self.binary_paths = {
                    "call": ["salt-call.exe"],
                    "cp": ["salt-cp.exe"],
                    "minion": ["salt-minion.exe"],
                    "pip": ["salt-pip.exe"],
                    "python": [python_bin],
                }
            else:
                if os.path.exists(self.install_dir / "bin" / "salt"):
                    install_dir = self.install_dir / "bin"
                else:
                    install_dir = self.install_dir
                self.binary_paths = {
                    "salt": [install_dir / "salt"],
                    "api": [install_dir / "salt-api"],
                    "call": [install_dir / "salt-call"],
                    "cloud": [install_dir / "salt-cloud"],
                    "cp": [install_dir / "salt-cp"],
                    "key": [install_dir / "salt-key"],
                    "master": [install_dir / "salt-master"],
                    "minion": [install_dir / "salt-minion"],
                    "proxy": [install_dir / "salt-proxy"],
                    "run": [install_dir / "salt-run"],
                    "ssh": [install_dir / "salt-ssh"],
                    "syndic": [install_dir / "salt-syndic"],
                    "spm": [install_dir / "spm"],
                    "pip": [install_dir / "salt-pip"],
                    "python": [python_bin],
                }
        else:
            if self.run_root and os.path.exists(self.run_root):
                if platform.is_windows():
                    self.binary_paths = {
                        "call": [str(self.run_root), "call"],
                        "cp": [str(self.run_root), "cp"],
                        "minion": [str(self.run_root), "minion"],
                        "pip": [str(self.run_root), "pip"],
                        "python": [python_bin],
                    }
                else:
                    self.binary_paths = {
                        "salt": [str(self.run_root)],
                        "api": [str(self.run_root), "api"],
                        "call": [str(self.run_root), "call"],
                        "cloud": [str(self.run_root), "cloud"],
                        "cp": [str(self.run_root), "cp"],
                        "key": [str(self.run_root), "key"],
                        "master": [str(self.run_root), "master"],
                        "minion": [str(self.run_root), "minion"],
                        "proxy": [str(self.run_root), "proxy"],
                        "run": [str(self.run_root), "run"],
                        "ssh": [str(self.run_root), "ssh"],
                        "syndic": [str(self.run_root), "syndic"],
                        "spm": [str(self.run_root), "spm"],
                        "pip": [str(self.run_root), "pip"],
                        "python": [python_bin],
                    }
            else:
                if platform.is_windows():
                    self.binary_paths = {
                        "call": [self.install_dir / "salt-call.exe"],
                        "cp": [self.install_dir / "salt-cp.exe"],
                        "minion": [self.install_dir / "salt-minion.exe"],
                        "pip": [self.install_dir / "salt-pip.exe"],
                        "python": [python_bin],
                    }
                else:
                    self.binary_paths = {
                        "salt": [self.install_dir / "salt"],
                        "api": [self.install_dir / "salt-api"],
                        "call": [self.install_dir / "salt-call"],
                        "cloud": [self.install_dir / "salt-cloud"],
                        "cp": [self.install_dir / "salt-cp"],
                        "key": [self.install_dir / "salt-key"],
                        "master": [self.install_dir / "salt-master"],
                        "minion": [self.install_dir / "salt-minion"],
                        "proxy": [self.install_dir / "salt-proxy"],
                        "run": [self.install_dir / "salt-run"],
                        "ssh": [self.install_dir / "salt-ssh"],
                        "syndic": [self.install_dir / "salt-syndic"],
                        "spm": [self.install_dir / "spm"],
                        "pip": [self.install_dir / "salt-pip"],
                        "python": [python_bin],
                    }

    @staticmethod
    def salt_factories_root_dir(system_service: bool = False) -> pathlib.Path:
        if system_service is False:
            return None
        if platform.is_windows():
            return pathlib.Path("C:/salt")
        if platform.is_darwin():
            return pathlib.Path("/opt/salt")
        return pathlib.Path("/")

    def _check_retcode(self, ret):
        """
        helper function ot check subprocess.run
        returncode equals 0, if not raise assertionerror
        """
        if ret.returncode != 0:
            log.error(ret)
        assert ret.returncode == 0
        return True

    @property
    def salt_hashes(self):
        for _hash in self.hashes.keys():
            for fpath in ARTIFACTS_DIR.glob(f"**/*{_hash}*"):
                fpath = str(fpath)
                if re.search(f"{_hash}", fpath):
                    self.hashes[_hash]["file"] = fpath

        return self.hashes

    def _install_ssm_service(self):
        # Register the services
        # run_root and ssm_bin are configured in helper.py to point to the
        # correct binary location
        log.debug("Installing master service")
        ret = self.proc.run(
            str(self.ssm_bin),
            "install",
            "salt-master",
            str(self.run_root),
            "master",
            "-c",
            str(self.conf_dir),
        )
        self._check_retcode(ret)
        log.debug("Installing minion service")
        ret = self.proc.run(
            str(self.ssm_bin),
            "install",
            "salt-minion",
            str(self.run_root),
            "minion",
            "-c",
            str(self.conf_dir),
        )
        self._check_retcode(ret)
        log.debug("Installing api service")
        ret = self.proc.run(
            str(self.ssm_bin),
            "install",
            "salt-api",
            str(self.run_root),
            "api",
            "-c",
            str(self.conf_dir),
        )
        self._check_retcode(ret)

    def _install_compressed(self, upgrade=False):
        pkg = self.pkgs[0]
        log.info("Installing %s", pkg)
        if platform.is_windows():
            if pkg.endswith("zip"):
                # Extract the files
                log.debug("Extracting zip file")
                with ZipFile(pkg, "r") as zip:
                    zip.extractall(path=self.root)
            elif pkg.endswith("exe") or pkg.endswith("msi"):
                log.error("Not a compressed package type: %s", pkg)
            else:
                log.error("Unknown package type: %s", pkg)
            if self.system_service:
                self._install_ssm_service()
        elif platform.is_darwin():
            log.debug("Extracting tarball into %s", self.root)
            with tarfile.open(pkg) as tar:  # , "r:gz")
                tar.extractall(path=str(self.root))
        else:
            log.debug("Extracting tarball into %s", self.root)
            with tarfile.open(pkg) as tar:  # , "r:gz")
                tar.extractall(path=str(self.root))

    def _install_pkgs(self, upgrade=False):
        pkg = self.pkgs[0]
        if platform.is_windows():
            if upgrade:
                self.root = self.install_dir.parent
                self.bin_dir = self.install_dir
                self.ssm_bin = self.install_dir / "ssm.exe"
            if pkg.endswith("exe"):
                # Install the package
                log.debug("Installing: %s", str(pkg))
                ret = self.proc.run(str(pkg), "/start-minion=0", "/S")
                self._check_retcode(ret)
            elif pkg.endswith("msi"):
                # Install the package
                log.debug("Installing: %s", str(pkg))
                # Write a batch file to run the installer. It is impossible to
                # perform escaping of the START_MINION property that the MSI
                # expects unless we do it via a batch file
                batch_file = pathlib.Path(pkg).parent / "install_msi.cmd"
                batch_content = f'msiexec /qn /i "{str(pkg)}" START_MINION=""\n'
                with open(batch_file, "w") as fp:
                    fp.write(batch_content)
                # Now run the batch file
                ret = self.proc.run("cmd.exe", "/c", str(batch_file))
                self._check_retcode(ret)
            else:
                log.error("Invalid package: %s", pkg)
                return False

            # Remove the service installed by the installer
            log.debug("Removing installed salt-minion service")
            self.proc.run(str(self.ssm_bin), "remove", "salt-minion", "confirm")
            self.update_process_path()

        elif platform.is_darwin():
            daemons_dir = pathlib.Path(os.sep, "Library", "LaunchDaemons")
            service_name = "com.saltstack.salt.minion"
            plist_file = daemons_dir / f"{service_name}.plist"
            log.debug("Installing: %s", str(pkg))
            ret = self.proc.run("installer", "-pkg", str(pkg), "-target", "/")
            self._check_retcode(ret)
            # Stop the service installed by the installer
            self.proc.run("launchctl", "disable", f"system/{service_name}")
            self.proc.run("launchctl", "bootout", "system", str(plist_file))
        elif upgrade:
            env = os.environ.copy()
            extra_args = []
            if self.distro_id in ("ubuntu", "debian"):
                env["DEBIAN_FRONTEND"] = "noninteractive"
                extra_args = [
                    "-o",
                    "DPkg::Options::=--force-confdef",
                    "-o",
                    "DPkg::Options::=--force-confold",
                ]
            log.info("Installing packages:\n%s", pprint.pformat(self.pkgs))
            args = extra_args + self.pkgs
            ret = self.proc.run(
                self.pkg_mngr,
                "upgrade",
                "-y",
                *args,
                _timeout=120,
                env=env,
            )
        else:
            log.info("Installing packages:\n%s", pprint.pformat(self.pkgs))
            ret = self.proc.run(self.pkg_mngr, "install", "-y", *self.pkgs)
        if not platform.is_darwin() and not platform.is_windows():
            # Make sure we don't have any trailing references to old package file locations
            ret.returncode == 0
            assert "/saltstack/salt/run" not in ret.stdout
        log.info(ret)
        self._check_retcode(ret)

    def install(self, upgrade=False):
        if self.compressed:
            self._install_compressed(upgrade=upgrade)
        else:
            self._install_pkgs(upgrade=upgrade)
            if self.distro_id in ("ubuntu", "debian"):
                self.stop_services()

    def stop_services(self):
        """
        Debian distros automatically start the services
        We want to ensure our tests start with the config
        settings we have set. This will also verify the expected
        services are up and running.
        """
        for service in ["salt-syndic", "salt-master", "salt-minion"]:
            check_run = self.proc.run("systemctl", "status", service)
            if check_run.returncode != 0:
                # The system was not started automatically and we
                # are expecting it to be on install
                log.debug("The service %s was not started on install.", service)
                return False
            stop_service = self.proc.run("systemctl", "stop", service)
            self._check_retcode(stop_service)
        return True

    def install_previous(self):
        """
        Install previous version. This is used for
        upgrade tests.
        """
        major_ver = self.major
        minor_ver = self.minor
        pkg_version = self.pkg_version
        full_version = f"{self.major}.{self.minor}-{pkg_version}"
        relenv = int(major_ver) >= 3006

        min_ver = f"{major_ver}"
        distro_name = self.distro_name
        if distro_name == "centos" or distro_name == "fedora":
            distro_name = "redhat"
        root_url = "salt/py3/"
        if self.classic:
            root_url = "py3/"

        if self.distro_name in ["redhat", "centos", "amazon", "fedora", "vmware"]:
            for fp in pathlib.Path("/etc", "yum.repos.d").glob("epel*"):
                fp.unlink()
            gpg_key = "SALTSTACK-GPG-KEY.pub"
            if self.distro_version == "9":
                gpg_key = "SALTSTACK-GPG-KEY2.pub"
            if relenv:
                gpg_key = "SALT-PROJECT-GPG-PUBKEY-2023.pub"

            if platform.is_aarch64():
                arch = "aarch64"
            else:
                arch = "x86_64"
            ret = self.proc.run(
                "rpm",
                "--import",
                f"https://repo.saltproject.io/{root_url}{distro_name}/{self.distro_version}/{arch}/{major_ver}/{gpg_key}",
            )
            self._check_retcode(ret)
            download_file(
                f"https://repo.saltproject.io/{root_url}{distro_name}/{self.distro_version}/{arch}/{major_ver}.repo",
                f"/etc/yum.repos.d/salt-{distro_name}.repo",
            )
            ret = self.proc.run(self.pkg_mngr, "clean", "expire-cache")
            self._check_retcode(ret)
            ret = self.proc.run(
                self.pkg_mngr,
                "install",
                *self.salt_pkgs,
                "-y",
            )
            self._check_retcode(ret)

        elif distro_name in ["debian", "ubuntu"]:
            ret = self.proc.run(self.pkg_mngr, "install", "curl", "-y")
            self._check_retcode(ret)
            ret = self.proc.run(self.pkg_mngr, "install", "apt-transport-https", "-y")
            self._check_retcode(ret)
            ## only classic 3005 has arm64 support
            if self.major >= "3006" and platform.is_aarch64():
                arch = "arm64"
            elif platform.is_aarch64() and self.classic:
                arch = "arm64"
            else:
                arch = "amd64"
            pathlib.Path("/etc/apt/keyrings").mkdir(parents=True, exist_ok=True)
            gpg_dest = "salt-archive-keyring.gpg"
            gpg_key = gpg_dest
            if relenv:
                gpg_key = "SALT-PROJECT-GPG-PUBKEY-2023.gpg"

            download_file(
                f"https://repo.saltproject.io/{root_url}{distro_name}/{self.distro_version}/{arch}/{major_ver}/{gpg_key}",
                f"/etc/apt/keyrings/{gpg_dest}",
            )
            with open(
                pathlib.Path("/etc", "apt", "sources.list.d", "salt.list"), "w"
            ) as fp:
                fp.write(
                    f"deb [signed-by=/etc/apt/keyrings/{gpg_dest} arch={arch}] "
                    f"https://repo.saltproject.io/{root_url}{distro_name}/{self.distro_version}/{arch}/{major_ver} {self.distro_codename} main"
                )
            ret = self.proc.run(self.pkg_mngr, "update")
            self._check_retcode(ret)
            ret = self.proc.run(
                self.pkg_mngr,
                "install",
                *self.salt_pkgs,
                "-y",
            )
            self._check_retcode(ret)
            self.stop_services()
        elif platform.is_windows():
            self.onedir = True
            self.installer_pkg = True
            self.bin_dir = self.install_dir / "bin"
            self.run_root = self.bin_dir / f"salt.exe"
            self.ssm_bin = self.bin_dir / "ssm.exe"
            if self.file_ext == "msi" or relenv:
                self.ssm_bin = self.install_dir / "ssm.exe"

            if not self.classic:
                if not relenv:
                    win_pkg = f"salt-{self.prev_version}-{pkg_version}-windows-amd64.{self.file_ext}"
                    win_pkg_url = f"https://repo.saltproject.io/salt/py3/windows/{major_ver}/{win_pkg}"
                else:
                    if self.file_ext == "msi":
                        win_pkg = (
                            f"Salt-Minion-{self.prev_version}-Py3-AMD64.{self.file_ext}"
                        )
                    elif self.file_ext == "exe":
                        win_pkg = f"Salt-Minion-{self.prev_version}-Py3-AMD64-Setup.{self.file_ext}"
                    win_pkg_url = f"https://repo.saltproject.io/salt/py3/windows/{major_ver}/{win_pkg}"
            else:
                if self.file_ext == "msi":
                    win_pkg = f"Salt-Minion-{min_ver}-1-Py3-AMD64.{self.file_ext}"
                elif self.file_ext == "exe":
                    win_pkg = f"Salt-Minion-{min_ver}-1-Py3-AMD64-Setup.{self.file_ext}"
                win_pkg_url = f"https://repo.saltproject.io/windows/{win_pkg}"
            pkg_path = pathlib.Path(r"C:\TEMP", win_pkg)
            pkg_path.parent.mkdir(exist_ok=True)
            download_file(win_pkg_url, pkg_path)

            if self.file_ext == "msi":
                # Write a batch file to run the installer. It is impossible to
                # perform escaping of the START_MINION property that the MSI
                # expects unless we do it via a batch file
                batch_file = pkg_path.parent / "install_msi.cmd"
                batch_content = f'msiexec /qn /i {str(pkg_path)} START_MINION=""'
                with open(batch_file, "w") as fp:
                    fp.write(batch_content)
                # Now run the batch file
                ret = self.proc.run("cmd.exe", "/c", str(batch_file))
                self._check_retcode(ret)
            else:
                ret = self.proc.run(pkg_path, "/start-minion=0", "/S")
                self._check_retcode(ret)

            log.debug("Removing installed salt-minion service")
            ret = self.proc.run(str(self.ssm_bin), "remove", "salt-minion", "confirm")
            self._check_retcode(ret)

            if self.system_service:
                self._install_system_service()

        elif platform.is_darwin():
            if self.classic:
                mac_pkg = f"salt-{self.prev_version}-py3-x86_64.pkg"
                mac_pkg_url = f"https://repo.saltproject.io/osx/{mac_pkg}"
            else:
                if not relenv:
                    mac_pkg = f"salt-{self.prev_version}-{pkg_version}-macos-x86_64.pkg"
                    mac_pkg_url = f"https://repo.saltproject.io/salt/py3/macos/{major_ver}/{mac_pkg}"
                else:
                    mac_pkg = f"salt-{self.prev_version}-py3-x86_64.pkg"
                    mac_pkg_url = f"https://repo.saltproject.io/salt/py3/macos/{major_ver}/{mac_pkg}"

            mac_pkg_path = f"/tmp/{mac_pkg}"
            if not os.path.exists(mac_pkg_path):
                download_file(
                    f"{mac_pkg_url}",
                    f"/tmp/{mac_pkg}",
                )

            ret = self.proc.run("installer", "-pkg", mac_pkg_path, "-target", "/")
            self._check_retcode(ret)

    def _uninstall_compressed(self):
        if platform.is_windows():
            if self.system_service:
                # Uninstall the services
                log.debug("Uninstalling master service")
                self.proc.run(str(self.ssm_bin), "stop", "salt-master")
                self.proc.run(str(self.ssm_bin), "remove", "salt-master", "confirm")
                log.debug("Uninstalling minion service")
                self.proc.run(str(self.ssm_bin), "stop", "salt-minion")
                self.proc.run(str(self.ssm_bin), "remove", "salt-minion", "confirm")
                log.debug("Uninstalling api service")
                self.proc.run(str(self.ssm_bin), "stop", "salt-api")
                self.proc.run(str(self.ssm_bin), "remove", "salt-api", "confirm")
            log.debug("Removing the Salt Service Manager")
            if self.ssm_bin:
                try:
                    self.ssm_bin.unlink()
                except PermissionError:
                    atexit.register(self.ssm_bin.unlink)
        if platform.is_darwin():
            # From here: https://stackoverflow.com/a/46118276/4581998
            daemons_dir = pathlib.Path(os.sep, "Library", "LaunchDaemons")
            for service in ("minion", "master", "api", "syndic"):
                service_name = f"com.saltstack.salt.{service}"
                plist_file = daemons_dir / f"{service_name}.plist"
                # Stop the services
                self.proc.run("launchctl", "disable", f"system/{service_name}")
                self.proc.run("launchctl", "bootout", "system", str(plist_file))

            # Remove Symlink to salt-config
            if os.path.exists("/usr/local/sbin/salt-config"):
                os.unlink("/usr/local/sbin/salt-config")

            # Remove supporting files
            self.proc.run(
                "pkgutil",
                "--only-files",
                "--files",
                "com.saltstack.salt",
                "|",
                "grep",
                "-v",
                "opt",
                "|",
                "tr",
                "'\n'",
                "' '",
                "|",
                "xargs",
                "-0",
                "rm",
                "-f",
            )

            # Remove directories
            if os.path.exists("/etc/salt"):
                shutil.rmtree("/etc/salt")

            # Remove path
            if os.path.exists("/etc/paths.d/salt"):
                os.remove("/etc/paths.d/salt")

            # Remove receipt
            self.proc.run("pkgutil", "--forget", "com.saltstack.salt")

        if self.singlebin:
            log.debug("Deleting the salt binary: %s", self.run_root)
            if self.run_root:
                try:
                    self.run_root.unlink()
                except PermissionError:
                    atexit.register(self.run_root.unlink)
        else:
            log.debug("Deleting the onedir directory: %s", self.root / "salt")
            shutil.rmtree(str(self.root / "salt"))

    def _uninstall_pkgs(self):
        pkg = self.pkgs[0]
        if platform.is_windows():
            log.info("Uninstalling %s", pkg)
            if pkg.endswith("exe"):
                uninst = self.install_dir / "uninst.exe"
                ret = self.proc.run(uninst, "/S")
                self._check_retcode(ret)
            elif pkg.endswith("msi"):
                ret = self.proc.run("msiexec.exe", "/qn", "/x", pkg)
                self._check_retcode(ret)

        elif platform.is_darwin():
            self._uninstall_compressed()
        else:
            log.debug("Un-Installing packages:\n%s", pprint.pformat(self.salt_pkgs))
            ret = self.proc.run(self.pkg_mngr, self.rm_pkg, "-y", *self.salt_pkgs)
            self._check_retcode(ret)

    def uninstall(self):
        if self.compressed:
            self._uninstall_compressed()
        else:
            self._uninstall_pkgs()

    def assert_uninstalled(self):
        """
        Assert that the paths in /opt/saltstack/ were correctly
        removed or not removed
        """
        return
        if platform.is_windows():
            # I'm not sure where the /opt/saltstack path is coming from
            # This is the path we're using to test windows
            opt_path = pathlib.Path(os.getenv("LocalAppData"), "salt", "pypath")
        else:
            opt_path = pathlib.Path(os.sep, "opt", "saltstack", "salt", "pypath")
        if not opt_path.exists():
            if platform.is_windows():
                assert not opt_path.parent.exists()
            else:
                assert not opt_path.parent.parent.exists()
        else:
            opt_path_contents = list(opt_path.rglob("*"))
            if not opt_path_contents:
                pytest.fail(
                    f"The path '{opt_path}' exists but there are no files in it."
                )
            else:
                for path in list(opt_path_contents):
                    if path.name in (".installs.json", "__pycache__"):
                        opt_path_contents.remove(path)
                if opt_path_contents:
                    pytest.fail(
                        "The test left some files behind: {}".format(
                            ", ".join([str(p) for p in opt_path_contents])
                        )
                    )

    def write_launchd_conf(self, service):
        service_name = f"com.saltstack.salt.{service}"
        ret = self.proc.run("launchctl", "list", service_name)
        # 113 means it couldn't find a service with that name
        if ret.returncode == 113:
            daemons_dir = pathlib.Path(os.sep, "Library", "LaunchDaemons")
            plist_file = daemons_dir / f"{service_name}.plist"
            # Make sure we're using this plist file
            if plist_file.exists():
                log.warning("Removing existing plist file for service: %s", service)
                plist_file.unlink()

            log.debug("Creating plist file for service: %s", service)
            contents = textwrap.dedent(
                f"""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
                <plist version="1.0">
                    <dict>
                        <key>Label</key>
                        <string>{service_name}</string>
                        <key>RunAtLoad</key>
                        <true/>
                        <key>KeepAlive</key>
                        <true/>
                        <key>ProgramArguments</key>
                        <array>
                            <string>{self.run_root}</string>
                            <string>{service}</string>
                            <string>-c</string>
                            <string>{self.conf_dir}</string>
                        </array>
                        <key>SoftResourceLimits</key>
                        <dict>
                            <key>NumberOfFiles</key>
                            <integer>100000</integer>
                        </dict>
                        <key>HardResourceLimits</key>
                        <dict>
                            <key>NumberOfFiles</key>
                            <integer>100000</integer>
                        </dict>
                    </dict>
                </plist>
                """
            )
            plist_file.write_text(contents, encoding="utf-8")
            contents = plist_file.read_text()
            log.debug("Created '%s'. Contents:\n%s", plist_file, contents)

            # Delete the plist file upon completion
            atexit.register(plist_file.unlink)

    def write_systemd_conf(self, service, binary):
        ret = self.proc.run("systemctl", "daemon-reload")
        self._check_retcode(ret)
        ret = self.proc.run("systemctl", "status", service)
        if ret.returncode in (3, 4):
            log.warning(
                "No systemd unit file was found for service %s. Creating one.", service
            )
            contents = textwrap.dedent(
                """\
                [Unit]
                Description={service}

                [Service]
                KillMode=process
                Type=notify
                NotifyAccess=all
                LimitNOFILE=8192
                ExecStart={tgt} -c {conf_dir}

                [Install]
                WantedBy=multi-user.target
                """
            )
            if isinstance(binary, list) and len(binary) == 1:
                binary = shutil.which(binary[0]) or binary[0]
            elif isinstance(binary, list):
                binary = " ".join(binary)
            unit_path = pathlib.Path(
                os.sep, "etc", "systemd", "system", f"{service}.service"
            )
            contents = contents.format(
                service=service, tgt=binary, conf_dir=self.conf_dir
            )
            log.info("Created '%s'. Contents:\n%s", unit_path, contents)
            unit_path.write_text(contents, encoding="utf-8")
            ret = self.proc.run("systemctl", "daemon-reload")
            atexit.register(unit_path.unlink)
            self._check_retcode(ret)

    def __enter__(self):
        if platform.is_windows():
            self.update_process_path()

        if not self.no_install:
            if self.upgrade:
                self.install_previous()
            else:
                self.install()
        return self

    def __exit__(self, *_):
        if not self.no_uninstall:
            self.uninstall()
            self.assert_uninstalled()


class PkgSystemdSaltDaemonImpl(SystemdSaltDaemonImpl):
    def get_service_name(self):
        if self._service_name is None:
            self._service_name = self.factory.script_name
        return self._service_name


@attr.s(kw_only=True)
class PkgLaunchdSaltDaemonImpl(PkgSystemdSaltDaemonImpl):

    plist_file = attr.ib()

    @plist_file.default
    def _default_plist_file(self):
        daemons_dir = pathlib.Path(os.sep, "Library", "LaunchDaemons")
        return daemons_dir / f"{self.get_service_name()}.plist"

    def get_service_name(self):
        if self._service_name is None:
            service_name = super().get_service_name()
            if "-" in service_name:
                service_name = service_name.split("-")[-1]
            self._service_name = f"com.saltstack.salt.{service_name}"
        return self._service_name

    def cmdline(self, *args):  # pylint: disable=arguments-differ
        """
        Construct a list of arguments to use when starting the subprocess.

        :param str args:
            Additional arguments to use when starting the subprocess

        """
        if args:  # pragma: no cover
            log.debug(
                "%s.run() is ignoring the passed in arguments: %r",
                self.__class__.__name__,
                args,
            )
        self._internal_run(
            "launchctl",
            "enable",
            f"system/{self.get_service_name()}",
        )
        return (
            "launchctl",
            "bootstrap",
            "system",
            str(self.plist_file),
        )

    def is_running(self):
        """
        Returns true if the sub-process is alive.
        """
        if self._process is None:
            ret = self._internal_run("launchctl", "list", self.get_service_name())
            if ret.stdout == "":
                return False

            if "PID" not in ret.stdout:
                return False

            pid = None
            # PID in a line that looks like this
            # "PID" = 445;
            for line in ret.stdout.splitlines():
                if "PID" in line:
                    pid = line.rstrip(";").split(" = ")[1]

            if pid is None:
                return False

            self._process = psutil.Process(int(pid))

        return self._process.is_running()

    def _terminate(self):
        """
        This method actually terminates the started daemon.
        """
        # We completely override the parent class method because we're not using
        # the self._terminal property, it's a launchd service
        if self._process is None:  # pragma: no cover
            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_result
            return (
                self._terminal_result
            )  # pylint: disable=access-member-before-definition

        atexit.unregister(self.terminate)
        log.info("Stopping %s", self.factory)
        pid = self.pid
        # Collect any child processes information before terminating the process
        with contextlib.suppress(psutil.NoSuchProcess):
            for child in psutil.Process(pid).children(recursive=True):
                if (
                    child not in self._children
                ):  # pylint: disable=access-member-before-definition
                    self._children.append(
                        child
                    )  # pylint: disable=access-member-before-definition

        if self._process.is_running():  # pragma: no cover
            cmdline = _get_cmdline(self._process)
        else:
            cmdline = []

        # Disable the service
        self._internal_run(
            "launchctl",
            "disable",
            f"system/{self.get_service_name()}",
        )
        # Unload the service
        self._internal_run("launchctl", "bootout", "system", str(self.plist_file))

        if self._process.is_running():  # pragma: no cover
            try:
                self._process.wait()
            except psutil.TimeoutExpired:
                self._process.terminate()
                try:
                    self._process.wait()
                except psutil.TimeoutExpired:
                    pass

        exitcode = self._process.wait() or 0

        # Dereference the internal _process attribute
        self._process = None
        # Lets log and kill any child processes left behind, including the main subprocess
        # if it failed to properly stop
        terminate_process(
            pid=pid,
            kill_children=True,
            children=self._children,  # pylint: disable=access-member-before-definition
            slow_stop=self.factory.slow_stop,
        )

        if self._terminal_stdout is not None:
            self._terminal_stdout.close()  # pylint: disable=access-member-before-definition
        if self._terminal_stderr is not None:
            self._terminal_stderr.close()  # pylint: disable=access-member-before-definition
        stdout = stderr = ""
        try:
            self._terminal_result = ProcessResult(
                returncode=exitcode, stdout=stdout, stderr=stderr, cmdline=cmdline
            )
            log.info("%s %s", self.factory.__class__.__name__, self._terminal_result)
            return self._terminal_result
        finally:
            self._terminal = None
            self._terminal_stdout = None
            self._terminal_stderr = None
            self._terminal_timeout = None
            self._children = []


@attr.s(kw_only=True)
class PkgSsmSaltDaemonImpl(PkgSystemdSaltDaemonImpl):
    def cmdline(self, *args):  # pylint: disable=arguments-differ
        """
        Construct a list of arguments to use when starting the subprocess.

        :param str args:
            Additional arguments to use when starting the subprocess

        """
        if args:  # pragma: no cover
            log.debug(
                "%s.run() is ignoring the passed in arguments: %r",
                self.__class__.__name__,
                args,
            )
        return (
            str(self.factory.salt_pkg_install.ssm_bin),
            "start",
            self.get_service_name(),
        )

    def is_running(self):
        """
        Returns true if the sub-process is alive.
        """
        if self._process is None:
            n = 1
            while True:
                if self._process is not None:
                    break
                time.sleep(1)
                ret = self._internal_run(
                    str(self.factory.salt_pkg_install.ssm_bin),
                    "processes",
                    self.get_service_name(),
                )
                log.warning(ret)
                if not ret.stdout or (ret.stdout and not ret.stdout.strip()):
                    if n >= 120:
                        return False
                    n += 1
                    continue
                for line in ret.stdout.splitlines():
                    log.warning("Line: %s", line)
                    if not line.strip():
                        continue
                    mainpid = line.strip().split()[0]
                    self._process = psutil.Process(int(mainpid))
                    break
        return self._process.is_running()

    def _terminate(self):
        """
        This method actually terminates the started daemon.
        """
        # We completely override the parent class method because we're not using the
        # self._terminal property, it's a systemd service
        if self._process is None:  # pragma: no cover
            if TYPE_CHECKING:
                # Make mypy happy
                assert self._terminal_result
            return (
                self._terminal_result
            )  # pylint: disable=access-member-before-definition

        atexit.unregister(self.terminate)
        log.info("Stopping %s", self.factory)
        pid = self.pid
        # Collect any child processes information before terminating the process
        with contextlib.suppress(psutil.NoSuchProcess):
            for child in psutil.Process(pid).children(recursive=True):
                if (
                    child not in self._children
                ):  # pylint: disable=access-member-before-definition
                    self._children.append(
                        child
                    )  # pylint: disable=access-member-before-definition

        if self._process.is_running():  # pragma: no cover
            cmdline = _get_cmdline(self._process)
        else:
            cmdline = []

        # Tell ssm to stop the service
        try:
            self._internal_run(
                str(self.factory.salt_pkg_install.ssm_bin),
                "stop",
                self.get_service_name(),
            )
        except FileNotFoundError:
            pass

        if self._process.is_running():  # pragma: no cover
            try:
                self._process.wait()
            except psutil.TimeoutExpired:
                self._process.terminate()
                try:
                    self._process.wait()
                except psutil.TimeoutExpired:
                    pass

        exitcode = self._process.wait() or 0

        # Dereference the internal _process attribute
        self._process = None
        # Lets log and kill any child processes left behind, including the main subprocess
        # if it failed to properly stop
        terminate_process(
            pid=pid,
            kill_children=True,
            children=self._children,  # pylint: disable=access-member-before-definition
            slow_stop=self.factory.slow_stop,
        )

        if self._terminal_stdout is not None:
            self._terminal_stdout.close()  # pylint: disable=access-member-before-definition
        if self._terminal_stderr is not None:
            self._terminal_stderr.close()  # pylint: disable=access-member-before-definition
        stdout = stderr = ""
        try:
            self._terminal_result = ProcessResult(
                returncode=exitcode, stdout=stdout, stderr=stderr, cmdline=cmdline
            )
            log.info("%s %s", self.factory.__class__.__name__, self._terminal_result)
            return self._terminal_result
        finally:
            self._terminal = None
            self._terminal_stdout = None
            self._terminal_stderr = None
            self._terminal_timeout = None
            self._children = []


@attr.s(kw_only=True)
class PkgMixin:
    salt_pkg_install: SaltPkgInstall = attr.ib()

    def get_script_path(self):
        if self.salt_pkg_install.compressed or (
            platform.is_darwin()
            and self.salt_pkg_install.classic
            and self.salt_pkg_install.upgrade
        ):
            if self.salt_pkg_install.run_root and os.path.exists(
                self.salt_pkg_install.run_root
            ):
                return str(self.salt_pkg_install.run_root)
            elif os.path.exists(self.salt_pkg_install.bin_dir / self.script_name):
                return str(self.salt_pkg_install.bin_dir / self.script_name)
            else:
                return str(self.salt_pkg_install.install_dir / self.script_name)
        return super().get_script_path()

    def get_base_script_args(self):
        base_script_args = []
        if self.salt_pkg_install.run_root and os.path.exists(
            self.salt_pkg_install.run_root
        ):
            if self.salt_pkg_install.compressed:
                if self.script_name == "spm":
                    base_script_args.append(self.script_name)
                elif self.script_name != "salt":
                    base_script_args.append(self.script_name.split("salt-")[-1])
        base_script_args.extend(super().get_base_script_args())
        return base_script_args

    def cmdline(self, *args, **kwargs):
        _cmdline = super().cmdline(*args, **kwargs)
        if self.salt_pkg_install.compressed is False:
            return _cmdline
        if _cmdline[0] == self.python_executable:
            _cmdline.pop(0)
        return _cmdline


@attr.s(kw_only=True)
class DaemonPkgMixin(PkgMixin):
    def __attrs_post_init__(self):
        if not platform.is_windows() and self.salt_pkg_install.system_service:
            if platform.is_darwin():
                self.write_launchd_conf()
            else:
                self.write_systemd_conf()

    def get_service_name(self):
        return self.script_name

    def write_launchd_conf(self):
        raise NotImplementedError

    def write_systemd_conf(self):
        raise NotImplementedError


@attr.s(kw_only=True)
class SaltMaster(DaemonPkgMixin, master.SaltMaster):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-master"
        master.SaltMaster.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_install and self.salt_pkg_install.system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("master")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-master", self.salt_pkg_install.binary_paths["master"]
        )

    def salt_minion_daemon(self, minion_id, **kwargs):
        return super().salt_minion_daemon(
            minion_id,
            factory_class=SaltMinion,
            salt_pkg_install=self.salt_pkg_install,
            **kwargs,
        )

    def salt_api_daemon(self, **kwargs):
        return super().salt_api_daemon(
            factory_class=SaltApi, salt_pkg_install=self.salt_pkg_install, **kwargs
        )

    def salt_key_cli(self, **factory_class_kwargs):
        return super().salt_key_cli(
            factory_class=SaltKey,
            salt_pkg_install=self.salt_pkg_install,
            **factory_class_kwargs,
        )

    def salt_cli(self, **factory_class_kwargs):
        return super().salt_cli(
            factory_class=SaltCli,
            salt_pkg_install=self.salt_pkg_install,
            **factory_class_kwargs,
        )


@attr.s(kw_only=True)
class SaltMasterWindows(SaltMaster):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self.script_name = cli_scripts.generate_script(
            bin_dir=self.factories_manager.scripts_dir,
            script_name="salt-master",
            code_dir=self.factories_manager.code_dir.parent,
        )

    def _get_impl_class(self):
        return DaemonImpl

    def cmdline(self, *args, **kwargs):
        cmdline_ = super().cmdline(*args, **kwargs)
        if self.python_executable:
            if cmdline_[0] != self.python_executable:
                cmdline_.insert(0, self.python_executable)
        return cmdline_


@attr.s(kw_only=True, slots=True)
class SaltMinion(DaemonPkgMixin, minion.SaltMinion):
    """
    Subclassed just to tweak the binary paths if needed and factory classes.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-minion"
        minion.SaltMinion.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_install and self.salt_pkg_install.system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("minion")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-minion", self.salt_pkg_install.binary_paths["minion"]
        )

    def salt_call_cli(self, **factory_class_kwargs):
        return super().salt_call_cli(
            factory_class=SaltCall,
            salt_pkg_install=self.salt_pkg_install,
            **factory_class_kwargs,
        )


@attr.s(kw_only=True, slots=True)
class SaltApi(DaemonPkgMixin, api.SaltApi):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-api"
        api.SaltApi.__attrs_post_init__(self)
        DaemonPkgMixin.__attrs_post_init__(self)

    def _get_impl_class(self):
        if self.system_install and self.salt_pkg_install.system_service:
            if platform.is_windows():
                return PkgSsmSaltDaemonImpl
            if platform.is_darwin():
                return PkgLaunchdSaltDaemonImpl
            return PkgSystemdSaltDaemonImpl
        return DaemonImpl

    def write_launchd_conf(self):
        self.salt_pkg_install.write_launchd_conf("api")

    def write_systemd_conf(self):
        self.salt_pkg_install.write_systemd_conf(
            "salt-api",
            self.salt_pkg_install.binary_paths["api"],
        )


@attr.s(kw_only=True, slots=True)
class SaltCall(PkgMixin, call.SaltCall):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        call.SaltCall.__attrs_post_init__(self)
        self.script_name = "salt-call"


@attr.s(kw_only=True, slots=True)
class SaltCli(PkgMixin, salt.SaltCli):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt"
        salt.SaltCli.__attrs_post_init__(self)


@attr.s(kw_only=True, slots=True)
class SaltKey(PkgMixin, key.SaltKey):
    """
    Subclassed just to tweak the binary paths if needed.
    """

    def __attrs_post_init__(self):
        self.script_name = "salt-key"
        key.SaltKey.__attrs_post_init__(self)


@attr.s(kw_only=True, slots=True)
class TestUser:
    """
    Add a test user
    """

    salt_call_cli = attr.ib()

    username = attr.ib(default="saltdev")
    # Must follow Windows Password Complexity requirements
    password = attr.ib(default="P@ssW0rd")
    _pw_record = attr.ib(init=False, repr=False, default=None)

    def salt_call_local(self, *args):
        ret = self.salt_call_cli.run("--local", *args)
        if ret.returncode != 0:
            log.error(ret)
        assert ret.returncode == 0
        return ret.data

    def add_user(self):
        log.debug("Adding system account %r", self.username)
        if platform.is_windows():
            self.salt_call_local("user.add", self.username, self.password)
        else:
            self.salt_call_local("user.add", self.username)
            hash_passwd = crypt.crypt(self.password, crypt.mksalt(crypt.METHOD_SHA512))
            self.salt_call_local("shadow.set_password", self.username, hash_passwd)
        assert self.username in self.salt_call_local("user.list_users")

    def remove_user(self):
        log.debug("Removing system account %r", self.username)
        if platform.is_windows():
            self.salt_call_local(
                "user.delete", self.username, "purge=True", "force=True"
            )
        else:
            self.salt_call_local("user.delete", self.username, "remove=True")

    @property
    def pw_record(self):
        if self._pw_record is None and HAS_PWD:
            self._pw_record = pwd.getpwnam(self.username)
        return self._pw_record

    @property
    def uid(self):
        if HAS_PWD:
            return self.pw_record.pw_uid
        return None

    @property
    def gid(self):
        if HAS_PWD:
            return self.pw_record.pw_gid
        return None

    @property
    def env(self):
        environ = os.environ.copy()
        environ["LOGNAME"] = environ["USER"] = self.username
        environ["HOME"] = self.pw_record.pw_dir
        return environ

    def __enter__(self):
        self.add_user()
        return self

    def __exit__(self, *_):
        self.remove_user()


@attr.s(kw_only=True, slots=True)
class ApiRequest:
    salt_api: SaltApi = attr.ib(repr=False)
    test_account: TestUser = attr.ib(repr=False)
    session: requests.Session = attr.ib(init=False, repr=False)
    api_uri: str = attr.ib(init=False)
    auth_data: Dict[str, str] = attr.ib(init=False)

    @session.default
    def _default_session(self):
        return requests.Session()

    @api_uri.default
    def _default_api_uri(self):
        return f"http://localhost:{self.salt_api.config['rest_cherrypy']['port']}"

    @auth_data.default
    def _default_auth_data(self):
        return {
            "username": self.test_account.username,
            "password": self.test_account.password,
            "eauth": "auto",
            "out": "json",
        }

    def post(self, url, data):
        post_data = dict(**self.auth_data, **data)
        resp = self.session.post(f"{self.api_uri}/run", data=post_data).json()
        minion = next(iter(resp["return"][0]))
        return resp["return"][0][minion]

    def __enter__(self):
        self.session.__enter__()
        return self

    def __exit__(self, *args):
        self.session.__exit__(*args)


@pytest.helpers.register
def remove_stale_minion_key(master, minion_id):
    key_path = os.path.join(master.config["pki_dir"], "minions", minion_id)
    if os.path.exists(key_path):
        os.unlink(key_path)
    else:
        log.debug("The minion(id=%r) key was not found at %s", minion_id, key_path)


@pytest.helpers.register
def remove_stale_master_key(master):
    keys_path = os.path.join(master.config["pki_dir"], "master")
    for key_name in ("master.pem", "master.pub"):
        key_path = os.path.join(keys_path, key_name)
        if os.path.exists(key_path):
            os.unlink(key_path)
        else:
            log.debug(
                "The master(id=%r) %s key was not found at %s",
                master.id,
                key_name,
                key_path,
            )
    key_path = os.path.join(master.config["pki_dir"], "minion", "minion_master.pub")
    if os.path.exists(key_path):
        os.unlink(key_path)
    else:
        log.debug(
            "The master(id=%r) minion_master.pub key was not found at %s",
            master.id,
            key_path,
        )


@pytest.helpers.register
def download_file(url, dest, auth=None):
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True, auth=auth) as r:
        r.raise_for_status()
        with open(dest, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
    return dest



"""
Tests for building and installing salt
"""
import json
import logging
import os
import pathlib
import re
import sys

import pytest

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.version
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_initial_onedir_failure,
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
]


def _check_skip(grains):
    if grains["os"] == "SUSE":
        return True
    return False


def use_static_requirements_ids(value):
    return "USE_STATIC_REQUIREMENTS={}".format("1" if value else "0")


@pytest.fixture(params=[True, False], ids=use_static_requirements_ids)
def use_static_requirements(request):
    return request.param


@pytest.fixture
def virtualenv(virtualenv, use_static_requirements):
    virtualenv.environ["USE_STATIC_REQUIREMENTS"] = (
        "1" if use_static_requirements else "0"
    )
    return virtualenv


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_wheel(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a bdist_wheel package
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)
        venv.run(
            venv.venv_python,
            "setup.py",
            "bdist_wheel",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.whl"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated wheel file")
        salt_generated_package = salt_generated_package[0]

        # Assert generate wheel version matches what salt reports as its version
        whl_ver = [
            x
            for x in salt_generated_package.name.split("-")
            if re.search(r"^\d.\d*", x)
        ][0]
        whl_ver_cmp = whl_ver.replace("_", "-")
        assert whl_ver_cmp == salt.version.__version__, "{} != {}".format(
            whl_ver_cmp, salt.version.__version__
        )

        # Because bdist_wheel supports pep517, we don't have to pre-install Salt's
        # dependencies before installing the wheel package
        if not use_static_requirements and salt.utils.platform.is_windows():
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")
        venv.install(str(salt_generated_package))

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.txt from the installed salt wheel
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
            "salt",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)

        installed_salt_path = pathlib.Path(venv.venv_dir)
        installed_salt_path = installed_salt_path.joinpath(*subdir)
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.txt"
        assert salt_generated_version_file_path.is_file()


def test_egg(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a bdist_egg package
    """
    # TODO: We should actually disallow generating an egg file
    # Let's create the testing virtualenv
    with virtualenv as venv:
        ret = venv.run(
            venv.venv_python,
            "-c",
            "import setuptools; print(setuptools.__version__)",
        )
        setuptools_version = ret.stdout.strip()
        ret = venv.run(venv.venv_python, "-m", "easy_install", "--version", check=False)
        if ret.returncode != 0:
            pytest.skip(
                "Setuptools version, {}, does not include the easy_install module".format(
                    setuptools_version
                )
            )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        # Looks like, at least on windows, setuptools also get's downloaded as a salt dependency.
        # Let's check and see if this newly installed version also has easy_install
        ret = venv.run(
            venv.venv_python,
            "-c",
            "import setuptools; print(setuptools.__version__)",
        )
        setuptools_version = ret.stdout.strip()
        ret = venv.run(venv.venv_python, "-m", "easy_install", "--version", check=False)
        if ret.returncode != 0:
            pytest.skip(
                "Setuptools version, {}, does not include the easy_install module".format(
                    setuptools_version
                )
            )

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "bdist_egg",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.egg"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated egg file")
        salt_generated_package = salt_generated_package[0]

        # Assert generate wheel version matches what salt reports as its version
        egg_ver = [
            x
            for x in salt_generated_package.name.split("-")
            if re.search(r"^\d.\d*", x)
        ][0]
        egg_ver_cmp = egg_ver.replace("_", "-")
        assert egg_ver_cmp == salt.version.__version__, "{} != {}".format(
            egg_ver_cmp, salt.version.__version__
        )

        # We cannot pip install an egg file, let's go old school
        venv.run(venv.venv_python, "-m", "easy_install", str(salt_generated_package))

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.txt from the installed salt egg
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)
        site_packages_dir = pathlib.Path(venv.venv_dir)
        site_packages_dir = site_packages_dir.joinpath(*subdir)
        assert site_packages_dir.is_dir()
        installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
        if not installed_salt_path:
            pytest.fail("Failed to find the installed salt path")
        log.debug("Installed salt path glob matches: %s", installed_salt_path)
        installed_salt_path = installed_salt_path[0] / "salt"
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.txt"
        assert salt_generated_version_file_path.is_file(), "{} is not a file".format(
            salt_generated_version_file_path
        )


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_sdist(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test building and installing a sdist package
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "sdist",
            "--dist-dir",
            str(cache_dir),
            cwd=src_dir,
        )

        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        salt_generated_package = list(cache_dir.glob("*.tar.gz"))
        if not salt_generated_package:
            pytest.fail("Could not find the generated sdist file")
        salt_generated_package = salt_generated_package[0]
        log.info("Generated sdist file: %s", salt_generated_package.name)

        # Assert generated sdist version matches what salt reports as its version
        sdist_ver_cmp = salt_generated_package.name.split(".tar.gz")[0].split("salt-")[
            -1
        ]
        assert sdist_ver_cmp == salt.version.__version__, "{} != {}".format(
            sdist_ver_cmp, salt.version.__version__
        )

        venv.install(str(salt_generated_package))

        # Let's also ensure we have a salt/_version.txt from the installed salt wheel
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
            "salt",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)

        installed_salt_path = pathlib.Path(venv.venv_dir)
        installed_salt_path = installed_salt_path.joinpath(*subdir)
        assert installed_salt_path.is_dir()
        salt_generated_version_file_path = installed_salt_path / "_version.txt"
        assert salt_generated_version_file_path.is_file()
        with salt_generated_version_file_path.open() as rfh:
            log.debug("_version.txt contents:\n >>>>>>\n%s\n <<<<<<", rfh.read())

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_setup_install(virtualenv, cache_dir, use_static_requirements, src_dir):
    """
    test installing directly from source
    """
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Setuptools installs pre-release packages if we don't pin to an exact version
        # Let's download and install requirements before, running salt's install test
        venv.run(
            venv.venv_python,
            "-m",
            "pip",
            "download",
            "--dest",
            str(cache_dir),
            src_dir,
        )
        packages = []
        for fname in cache_dir.iterdir():
            if (
                fname.name.startswith("pycurl")
                and salt.utils.platform.is_windows()
                and not use_static_requirements
            ):
                # On windows, the latest pycurl release, 7.43.0.6 at the time of writing,
                # does not have wheel files uploaded, so, delete the downloaded source
                # tarball and will later force pycurl==7.43.0.5 to be pre-installed before
                # installing salt
                fname.unlink()
                continue
            packages.append(fname)
        venv.install(*[str(pkg) for pkg in packages])
        for package in packages:
            package.unlink()

        if salt.utils.platform.is_windows() and not use_static_requirements:
            # Like mentioned above, install pycurl==7.43.0.5
            # However, on windows, the latest pycurl release, 7.43.0.6 at the time of writing,
            # does not have wheel files uploaded, so, we force pycurl==7.43.0.5 to be
            # pre-installed before installing salt
            venv.install("pycurl==7.43.0.5")

        venv.run(
            venv.venv_python,
            "setup.py",
            "install",
            "--prefix",
            str(venv.venv_dir),
            cwd=src_dir,
        )

        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)

        # Let's ensure the version is correct
        cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
        for details in json.loads(cmd.stdout):
            if details["name"] != "salt":
                continue
            installed_version = details["version"]
            break
        else:
            pytest.fail("Salt was not found installed")

        # Let's compare the installed version with the version salt reports
        assert installed_version == salt.version.__version__, "{} != {}".format(
            installed_version, salt.version.__version__
        )

        # Let's also ensure we have a salt/_version.txt from the installed salt
        subdir = [
            "lib",
            "python{}.{}".format(*sys.version_info),
            "site-packages",
        ]
        if salt.utils.platform.is_windows():
            subdir.pop(1)
        site_packages_dir = pathlib.Path(venv.venv_dir)
        site_packages_dir = site_packages_dir.joinpath(*subdir)
        assert site_packages_dir.is_dir()
        installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
        if not installed_salt_path:
            pytest.fail("Failed to find the installed salt path")
        installed_salt_path = installed_salt_path[0] / "salt"
        salt_generated_version_file_path = installed_salt_path / "_version.txt"
        assert salt_generated_version_file_path.is_file()


def test_salt_install_args(
    virtualenv, cache_dir, use_static_requirements, src_dir, tmp_path
):
    """
    test building with `install` command with --salt-*
    args. For example, --salt-config-dir and --salt-cache-dir.
    """
    cache_dir = tmp_path / "cache_dir"
    config_dir = tmp_path / "config_dir"
    # Let's create the testing virtualenv
    with virtualenv as venv:
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)
        env = os.environ.copy()
        env["GENERATE_SALT_SYSPATHS"] = "True"
        ret = venv.run(
            venv.venv_python,
            "setup.py",
            "--salt-config-dir",
            str(config_dir),
            "--salt-cache-dir",
            str(cache_dir),
            "install",
            cwd=src_dir,
            env=env,
        )
        assert ret.returncode == 0
        syspath = pathlib.Path(src_dir, "build", "lib", "salt", "_syspaths.py")
        assert syspath.exists()
        with salt.utils.files.fopen(syspath) as fp:
            data = fp.read()
        assert str(cache_dir) in data
        assert str(config_dir) in data
        venv.run(venv.venv_python, "setup.py", "clean", cwd=src_dir)



import logging
import subprocess

import pytest

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_pip_install(install_salt, salt_call_cli):
    """
    Test pip.install and ensure that a package included in the tiamat build can be upgraded
    """
    ret = subprocess.run(
        install_salt.binary_paths["salt"] + ["--versions-report"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        check=True,
        shell=False,
    )
    assert ret.returncode == 0

    possible_upgrades = [
        "docker-py",
        "msgpack",
        "pycparser",
        "python-gnupg",
        "pyyaml",
        "pyzmq",
        "jinja2",
    ]
    found_new = False
    for dep in possible_upgrades:
        get_latest = salt_call_cli.run("--local", "pip.list_all_versions", dep)
        if not get_latest.data:
            # No information available
            continue
        dep_version = get_latest.data[-1]
        installed_version = None
        for line in ret.stdout.splitlines():
            if dep in line.lower():
                installed_version = line.lower().strip().split(":")[-1].strip()
                break
        else:
            pytest.fail(f"Failed to find {dep} in the versions report output")

        if dep_version == installed_version:
            log.warning(f"The {dep} dependency is already latest")
        else:
            found_new = True
            break

    if found_new:
        try:
            install = salt_call_cli.run(
                "--local", "pip.install", f"{dep}=={dep_version}"
            )
            assert install
            log.warning(install)
            # The assert is commented out because pip will actually trigger a failure since
            # we're breaking the dependency tree, but, for the purpose of this test, we can
            # ignore it.
            #
            # assert install.returncode == 0

            ret = subprocess.run(
                install_salt.binary_paths["salt"] + ["--versions-report"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                check=True,
                shell=False,
            )
            assert ret.returncode == 0
            for line in ret.stdout.splitlines():
                if dep in line.lower():
                    new_version = line.lower().strip().split(":")[-1].strip()
                    if new_version == installed_version:
                        pytest.fail(
                            f"The newly installed version of {dep} does not show in the versions report"
                        )
                    assert new_version == dep_version
                    break
            else:
                pytest.fail(f"Failed to find {dep} in the versions report output")
        finally:
            log.info(f"Uninstalling {dep_version}")
            assert salt_call_cli.run(
                "--local", "pip.uninstall", f"{dep}=={dep_version}"
            )
    else:
        pytest.skip("Did not find an upgrade version for any of the dependencies")



"""
Connection module for Amazon Route53

.. versionadded:: 2014.7.0

:configuration: This module accepts explicit route53 credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: yaml

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        route53.keyid: GKTADJGHEIQSXMKKRBJ08H
        route53.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        route53.region: us-east-1

    If a region is not specified, the default is 'universal', which is what the boto_route53
    library expects, rather than None.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
          keyid: GKTADJGHEIQSXMKKRBJ08H
          key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
          region: us-east-1

:depends: boto
"""
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging
import time

import salt.utils.compat
import salt.utils.odict as odict
import salt.utils.versions
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

try:
    # pylint: disable=unused-import
    import boto
    import boto.route53
    import boto.route53.healthcheck
    from boto.route53.exception import DNSServerError

    # pylint: enable=unused-import
    logging.getLogger("boto").setLevel(logging.CRITICAL)
    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    # create_zone params were changed in boto 2.35+
    return salt.utils.versions.check_boto_reqs(boto_ver="2.35.0", check_boto3=False)


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto.assign_funcs"](__name__, "route53", pack=__salt__)


def _get_split_zone(zone, _conn, private_zone):
    """
    With boto route53, zones can only be matched by name
    or iterated over in a list.  Since the name will be the
    same for public and private zones in a split DNS situation,
    iterate over the list and match the zone name and public/private
    status.
    """
    for _zone in _conn.get_zones():
        if _zone.name == zone:
            _private_zone = (
                True if _zone.config["PrivateZone"].lower() == "true" else False
            )
            if _private_zone == private_zone:
                return _zone
    return False


def _is_retryable_error(exception):
    return exception.code not in ["SignatureDoesNotMatch"]


def describe_hosted_zones(
    zone_id=None, domain_name=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return detailed info about one, or all, zones in the bound account.
    If neither zone_id nor domain_name is provided, return all zones.
    Note that the return format is slightly different between the 'all'
    and 'single' description types.

    zone_id
        The unique identifier for the Hosted Zone

    domain_name
        The FQDN of the Hosted Zone (including final period)

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.describe_hosted_zones domain_name=foo.bar.com. \
                profile='{"region": "us-east-1", "keyid": "A12345678AB", "key": "xblahblahblah"}'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if zone_id and domain_name:
        raise SaltInvocationError(
            "At most one of zone_id or domain_name may be provided"
        )
    retries = 10
    while retries:
        try:
            if zone_id:
                zone_id = (
                    zone_id.replace("/hostedzone/", "")
                    if zone_id.startswith("/hostedzone/")
                    else zone_id
                )
                ret = getattr(
                    conn.get_hosted_zone(zone_id), "GetHostedZoneResponse", None
                )
            elif domain_name:
                ret = getattr(
                    conn.get_hosted_zone_by_name(domain_name),
                    "GetHostedZoneResponse",
                    None,
                )
            else:
                marker = None
                ret = None
                while marker != "":
                    r = conn.get_all_hosted_zones(start_marker=marker, zone_list=ret)
                    ret = r["ListHostedZonesResponse"]["HostedZones"]
                    marker = r["ListHostedZonesResponse"].get("NextMarker", "")
            return ret if ret else []
        except DNSServerError as e:
            if retries:
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                retries -= 1
                continue
            log.error("Could not list zones: %s", e.message)
            return []


def list_all_zones_by_name(region=None, key=None, keyid=None, profile=None):
    """
    List, by their FQDNs, all hosted zones in the bound account.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.list_all_zones_by_name
    """
    ret = describe_hosted_zones(region=region, key=key, keyid=keyid, profile=profile)
    return [r["Name"] for r in ret]


def list_all_zones_by_id(region=None, key=None, keyid=None, profile=None):
    """
    List, by their IDs, all hosted zones in the bound account.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.list_all_zones_by_id
    """
    ret = describe_hosted_zones(region=region, key=key, keyid=keyid, profile=profile)
    return [r["Id"].replace("/hostedzone/", "") for r in ret]


def zone_exists(
    zone,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    retry_on_rate_limit=None,
    rate_limit_retries=None,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Check for the existence of a Route53 hosted zone.

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.zone_exists example.org

    retry_on_errors
        Continue to query if the zone exists after an error is
        raised. The previously used argument `retry_on_rate_limit`
        was deprecated for this argument. Users can still use
        `retry_on_rate_limit` to ensure backwards compatibility,
        but please migrate to using the favored `retry_on_errors`
        argument instead.

    error_retries
        Number of times to attempt to query if the zone exists.
        The previously used argument `rate_limit_retries` was
        deprecated for this arguments. Users can still use
        `rate_limit_retries` to ensure backwards compatibility,
        but please migrate to using the favored `error_retries`
        argument instead.
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if retry_on_rate_limit or rate_limit_retries is not None:
        if retry_on_rate_limit is not None:
            retry_on_errors = retry_on_rate_limit
        if rate_limit_retries is not None:
            error_retries = rate_limit_retries

    while error_retries > 0:
        try:
            return bool(conn.get_zone(zone))

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request "
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise
    return False


def create_zone(
    zone,
    private=False,
    vpc_id=None,
    vpc_region=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create a Route53 hosted zone.

    .. versionadded:: 2015.8.0

    zone
        DNS zone to create

    private
        True/False if the zone will be a private zone

    vpc_id
        VPC ID to associate the zone to (required if private is True)

    vpc_region
        VPC Region (required if private is True)

    region
        region endpoint to connect to

    key
        AWS key

    keyid
        AWS keyid

    profile
        AWS pillar profile

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.create_zone example.org
    """
    if region is None:
        region = "universal"

    if private:
        if not vpc_id or not vpc_region:
            msg = "vpc_id and vpc_region must be specified for a private zone"
            raise SaltInvocationError(msg)

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    _zone = conn.get_zone(zone)

    if _zone:
        return False

    conn.create_zone(zone, private_zone=private, vpc_id=vpc_id, vpc_region=vpc_region)
    return True


def create_healthcheck(
    ip_addr=None,
    fqdn=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    port=53,
    hc_type="TCP",
    resource_path="",
    string_match=None,
    request_interval=30,
    failure_threshold=3,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Create a Route53 healthcheck

    .. versionadded:: 2018.3.0

    ip_addr

        IP address to check.  ip_addr or fqdn is required.

    fqdn

        Domain name of the endpoint to check.  ip_addr or fqdn is required

    port

        Port to check

    hc_type

        Healthcheck type.  HTTP | HTTPS | HTTP_STR_MATCH | HTTPS_STR_MATCH | TCP

    resource_path

        Path to check

    string_match

        If hc_type is HTTP_STR_MATCH or HTTPS_STR_MATCH, the string to search for in the
        response body from the specified resource

    request_interval

        The number of seconds between the time that Amazon Route 53 gets a response from
        your endpoint and the time that it sends the next health-check request.

    failure_threshold

        The number of consecutive health checks that an endpoint must pass or fail for
        Amazon Route 53 to change the current status of the endpoint from unhealthy to
        healthy or vice versa.

    region

        Region endpoint to connect to

    key

        AWS key

    keyid

        AWS keyid

    profile

        AWS pillar profile

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.create_healthcheck 192.168.0.1
        salt myminion boto_route53.create_healthcheck 192.168.0.1 port=443 hc_type=HTTPS \
                                                      resource_path=/ fqdn=blog.saltstack.furniture
    """
    if fqdn is None and ip_addr is None:
        msg = "One of the following must be specified: fqdn or ip_addr"
        log.error(msg)
        return {"error": msg}
    hc_ = boto.route53.healthcheck.HealthCheck(
        ip_addr,
        port,
        hc_type,
        resource_path,
        fqdn=fqdn,
        string_match=string_match,
        request_interval=request_interval,
        failure_threshold=failure_threshold,
    )

    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    while error_retries > 0:
        try:
            return {"result": conn.create_health_check(hc_)}
        except DNSServerError as exc:
            log.debug(exc)
            if retry_on_errors and _is_retryable_error(exc):
                if "Throttling" == exc.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == exc.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            return {"error": __utils__["boto.get_error"](exc)}
    return False


def delete_zone(zone, region=None, key=None, keyid=None, profile=None):
    """
    Delete a Route53 hosted zone.

    .. versionadded:: 2015.8.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.delete_zone example.org
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    _zone = conn.get_zone(zone)

    if _zone:
        conn.delete_hosted_zone(_zone.id)
        return True
    return False


def _encode_name(name):
    return name.replace("*", r"\052")


def _decode_name(name):
    return name.replace(r"\052", "*")


def get_record(
    name,
    zone,
    record_type,
    fetch_all=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    split_dns=False,
    private_zone=False,
    identifier=None,
    retry_on_rate_limit=None,
    rate_limit_retries=None,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Get a record from a zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.get_record test.example.org example.org A

    retry_on_errors
        Continue to query if the zone exists after an error is
        raised. The previously used argument `retry_on_rate_limit`
        was deprecated for this argument. Users can still use
        `retry_on_rate_limit` to ensure backwards compatibility,
        but please migrate to using the favored `retry_on_errors`
        argument instead.

    error_retries
        Number of times to attempt to query if the zone exists.
        The previously used argument `rate_limit_retries` was
        deprecated for this arguments. Users can still use
        `rate_limit_retries` to ensure backwards compatibility,
        but please migrate to using the favored `error_retries`
        argument instead.
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if retry_on_rate_limit or rate_limit_retries is not None:
        if retry_on_rate_limit is not None:
            retry_on_errors = retry_on_rate_limit
        if rate_limit_retries is not None:
            error_retries = rate_limit_retries

    _record = None
    ret = odict.OrderedDict()
    while error_retries > 0:
        try:
            if split_dns:
                _zone = _get_split_zone(zone, conn, private_zone)
            else:
                _zone = conn.get_zone(zone)
            if not _zone:
                msg = "Failed to retrieve zone {}".format(zone)
                log.error(msg)
                return None
            _type = record_type.upper()

            name = _encode_name(name)

            _record = _zone.find_records(
                name, _type, all=fetch_all, identifier=identifier
            )

            break  # the while True

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise

    if _record:
        ret["name"] = _decode_name(_record.name)
        ret["value"] = _record.resource_records[0]
        ret["record_type"] = _record.type
        ret["ttl"] = _record.ttl
        if _record.identifier:
            ret["identifier"] = []
            ret["identifier"].append(_record.identifier)
            ret["identifier"].append(_record.weight)

    return ret


def _munge_value(value, _type):
    split_types = ["A", "MX", "AAAA", "TXT", "SRV", "SPF", "NS"]
    if _type in split_types:
        return value.split(",")
    return value


def add_record(
    name,
    value,
    zone,
    record_type,
    identifier=None,
    ttl=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    wait_for_sync=True,
    split_dns=False,
    private_zone=False,
    retry_on_rate_limit=None,
    rate_limit_retries=None,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Add a record to a zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.add_record test.example.org 1.1.1.1 example.org A

    retry_on_errors
        Continue to query if the zone exists after an error is
        raised. The previously used argument `retry_on_rate_limit`
        was deprecated for this argument. Users can still use
        `retry_on_rate_limit` to ensure backwards compatibility,
        but please migrate to using the favored `retry_on_errors`
        argument instead.

    error_retries
        Number of times to attempt to query if the zone exists.
        The previously used argument `rate_limit_retries` was
        deprecated for this arguments. Users can still use
        `rate_limit_retries` to ensure backwards compatibility,
        but please migrate to using the favored `error_retries`
        argument instead.
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if retry_on_rate_limit or rate_limit_retries is not None:
        if retry_on_rate_limit is not None:
            retry_on_errors = retry_on_rate_limit
        if rate_limit_retries is not None:
            error_retries = rate_limit_retries

    while error_retries > 0:
        try:
            if split_dns:
                _zone = _get_split_zone(zone, conn, private_zone)
            else:
                _zone = conn.get_zone(zone)
            if not _zone:
                msg = "Failed to retrieve zone {}".format(zone)
                log.error(msg)
                return False
            _type = record_type.upper()
            break

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise

    _value = _munge_value(value, _type)
    while error_retries > 0:
        try:
            # add_record requires a ttl value, annoyingly.
            if ttl is None:
                ttl = 60
            status = _zone.add_record(_type, name, _value, ttl, identifier)
            return _wait_for_sync(status.id, conn, wait_for_sync)

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise
    return False


def update_record(
    name,
    value,
    zone,
    record_type,
    identifier=None,
    ttl=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    wait_for_sync=True,
    split_dns=False,
    private_zone=False,
    retry_on_rate_limit=None,
    rate_limit_retries=None,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Modify a record in a zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.modify_record test.example.org 1.1.1.1 example.org A

    retry_on_errors
        Continue to query if the zone exists after an error is
        raised. The previously used argument `retry_on_rate_limit`
        was deprecated for this argument. Users can still use
        `retry_on_rate_limit` to ensure backwards compatibility,
        but please migrate to using the favored `retry_on_errors`
        argument instead.

    error_retries
        Number of times to attempt to query if the zone exists.
        The previously used argument `rate_limit_retries` was
        deprecated for this arguments. Users can still use
        `rate_limit_retries` to ensure backwards compatibility,
        but please migrate to using the favored `error_retries`
        argument instead.
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if split_dns:
        _zone = _get_split_zone(zone, conn, private_zone)
    else:
        _zone = conn.get_zone(zone)
    if not _zone:
        msg = "Failed to retrieve zone {}".format(zone)
        log.error(msg)
        return False
    _type = record_type.upper()

    if retry_on_rate_limit or rate_limit_retries is not None:
        if retry_on_rate_limit is not None:
            retry_on_errors = retry_on_rate_limit
        if rate_limit_retries is not None:
            error_retries = rate_limit_retries

    _value = _munge_value(value, _type)
    while error_retries > 0:
        try:
            old_record = _zone.find_records(name, _type, identifier=identifier)
            if not old_record:
                return False
            status = _zone.update_record(old_record, _value, ttl, identifier)
            return _wait_for_sync(status.id, conn, wait_for_sync)

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise
    return False


def delete_record(
    name,
    zone,
    record_type,
    identifier=None,
    all_records=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    wait_for_sync=True,
    split_dns=False,
    private_zone=False,
    retry_on_rate_limit=None,
    rate_limit_retries=None,
    retry_on_errors=True,
    error_retries=5,
):
    """
    Modify a record in a zone.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.delete_record test.example.org example.org A

    retry_on_errors
        Continue to query if the zone exists after an error is
        raised. The previously used argument `retry_on_rate_limit`
        was deprecated for this argument. Users can still use
        `retry_on_rate_limit` to ensure backwards compatibility,
        but please migrate to using the favored `retry_on_errors`
        argument instead.

    error_retries
        Number of times to attempt to query if the zone exists.
        The previously used argument `rate_limit_retries` was
        deprecated for this arguments. Users can still use
        `rate_limit_retries` to ensure backwards compatibility,
        but please migrate to using the favored `error_retries`
        argument instead.
    """
    if region is None:
        region = "universal"

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if split_dns:
        _zone = _get_split_zone(zone, conn, private_zone)
    else:
        _zone = conn.get_zone(zone)
    if not _zone:
        msg = "Failed to retrieve zone {}".format(zone)
        log.error(msg)
        return False
    _type = record_type.upper()

    if retry_on_rate_limit or rate_limit_retries is not None:
        if retry_on_rate_limit is not None:
            retry_on_errors = retry_on_rate_limit
        if rate_limit_retries is not None:
            error_retries = rate_limit_retries

    while error_retries > 0:
        try:
            old_record = _zone.find_records(
                name, _type, all=all_records, identifier=identifier
            )
            if not old_record:
                return False
            status = _zone.delete_record(old_record)
            return _wait_for_sync(status.id, conn, wait_for_sync)

        except DNSServerError as e:
            if retry_on_errors and _is_retryable_error(e):
                if "Throttling" == e.code:
                    log.debug("Throttled by AWS API.")
                elif "PriorRequestNotComplete" == e.code:
                    log.debug(
                        "The request was rejected by AWS API. "
                        "Route 53 was still processing a prior request."
                    )
                time.sleep(3)
                error_retries -= 1
                continue
            raise


def _try_func(conn, func, **args):
    tries = 30
    while True:
        try:
            return getattr(conn, func)(**args)
        except AttributeError as e:
            # Don't include **args in log messages - security concern.
            log.error(
                "Function `%s()` not found for AWS connection object %s", func, conn
            )
            return None
        except DNSServerError as e:
            if tries and e.code == "Throttling":
                log.debug("Throttled by AWS API.  Will retry in 5 seconds")
                time.sleep(5)
                tries -= 1
                continue
            log.error("Failed calling %s(): %s", func, e)
            return None


def _wait_for_sync(status, conn, wait=True):
    ### Wait should be a bool or an integer
    if wait is True:
        wait = 600
    if not wait:
        return True
    orig_wait = wait
    log.info("Waiting up to %s seconds for Route53 changes to synchronize", orig_wait)
    while wait > 0:
        change = conn.get_change(status)
        current = change.GetChangeResponse.ChangeInfo.Status
        if current == "INSYNC":
            return True
        sleep = wait if wait % 60 == wait else 60
        log.info(
            "Sleeping %s seconds waiting for changes to synch (current status %s)",
            sleep,
            current,
        )
        time.sleep(sleep)
        wait -= sleep
        continue
    log.error("Route53 changes not synced after %s seconds.", orig_wait)
    return False


def create_hosted_zone(
    domain_name,
    caller_ref=None,
    comment="",
    private_zone=False,
    vpc_id=None,
    vpc_name=None,
    vpc_region=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create a new Route53 Hosted Zone. Returns a Python data structure with information about the
    newly created Hosted Zone.

    domain_name
        The name of the domain. This must be fully-qualified, terminating with a period.  This is
        the name you have registered with your domain registrar.  It is also the name you will
        delegate from your registrar to the Amazon Route 53 delegation servers returned in response
        to this request.

    caller_ref
        A unique string that identifies the request and that allows create_hosted_zone() calls to
        be retried without the risk of executing the operation twice.  It can take several minutes
        for the change to replicate globally, and change from PENDING to INSYNC status. Thus it's
        best to provide some value for this where possible, since duplicate calls while the first
        is in PENDING status will be accepted and can lead to multiple copies of the zone being
        created.  On the other hand, if a zone is created with a given caller_ref, then deleted,
        a second attempt to create a zone with the same caller_ref will fail until that caller_ref
        is flushed from the Route53 system, which can take upwards of 24 hours.

    comment
        Any comments you want to include about the hosted zone.

    private_zone
        Set True if creating a private hosted zone.

    vpc_id
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with vpe_name.  Ignored when creating a non-private zone.

    vpc_name
        When creating a private hosted zone, either the VPC ID or VPC Name to associate with is
        required.  Exclusive with vpe_id.  Ignored when creating a non-private zone.

    vpc_region
        When creating a private hosted zone, the region of the associated VPC is required.  If not
        provided, an effort will be made to determine it from vpc_id or vpc_name, where possible.
        If this fails, you'll need to provide an explicit value for this option.  Ignored when
        creating a non-private zone.

    region
        Region endpoint to connect to.

    key
        AWS key to bind with.

    keyid
        AWS keyid to bind with.

    profile
        Dict, or pillar key pointing to a dict, containing AWS region/key/keyid.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_route53.create_hosted_zone example.org
    """
    if region is None:
        region = "universal"

    if not domain_name.endswith("."):
        raise SaltInvocationError(
            "Domain MUST be fully-qualified, complete with ending period."
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    deets = conn.get_hosted_zone_by_name(domain_name)
    if deets:
        log.info("Route53 hosted zone %s already exists", domain_name)
        return None

    args = {
        "domain_name": domain_name,
        "caller_ref": caller_ref,
        "comment": comment,
        "private_zone": private_zone,
    }

    if private_zone:
        if not _exactly_one((vpc_name, vpc_id)):
            raise SaltInvocationError(
                "Either vpc_name or vpc_id is required when creating a private zone."
            )
        vpcs = __salt__["boto_vpc.describe_vpcs"](
            vpc_id=vpc_id,
            name=vpc_name,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ).get("vpcs", [])
        if vpc_region and vpcs:
            vpcs = [v for v in vpcs if v["region"] == vpc_region]
        if not vpcs:
            log.error(
                "Private zone requested but a VPC matching given criteria not found."
            )
            return None
        if len(vpcs) > 1:
            log.error(
                "Private zone requested but multiple VPCs matching given "
                "criteria found: %s.",
                [v["id"] for v in vpcs],
            )
            return None
        vpc = vpcs[0]
        if vpc_name:
            vpc_id = vpc["id"]
        if not vpc_region:
            vpc_region = vpc["region"]
        args.update({"vpc_id": vpc_id, "vpc_region": vpc_region})
    else:
        if any((vpc_id, vpc_name, vpc_region)):
            log.info(
                "Options vpc_id, vpc_name, and vpc_region are ignored "
                "when creating non-private zones."
            )

    r = _try_func(conn, "create_hosted_zone", **args)
    if r is None:
        log.error("Failed to create hosted zone %s", domain_name)
        return None
    r = r.get("CreateHostedZoneResponse", {})
    # Pop it since it'll be irrelevant by the time we return
    status = r.pop("ChangeInfo", {}).get("Id", "").replace("/change/", "")
    synced = _wait_for_sync(status, conn, wait=600)
    if not synced:
        log.error("Hosted zone %s not synced after 600 seconds.", domain_name)
        return None
    return r



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.mark.slow_test
def test_sync_grains(salt_call_cli):
    ret = salt_call_cli.run("saltutil.sync_grains")
    assert ret.returncode == 0
    try:
        assert ret.data == []
    except AssertionError:
        # Maybe it had to sync again on the above call.
        # On this next call though, it shold return an empty list
        ret = salt_call_cli.run("saltutil.sync_grains")
        assert ret.returncode == 0
        assert ret.data == []



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest

import salt.states.boto_sns as boto_sns
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_sns: {}}


def test_present():
    """
    Test to ensure the SNS topic exists.
    """
    name = "test.example.com."

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False, False])
    mock_bool = MagicMock(return_value=False)
    with patch.dict(
        boto_sns.__salt__, {"boto_sns.exists": mock, "boto_sns.create": mock_bool}
    ):
        comt = "AWS SNS topic {} present.".format(name)
        ret.update({"comment": comt})
        assert boto_sns.present(name) == ret

        with patch.dict(boto_sns.__opts__, {"test": True}):
            comt = "AWS SNS topic {} is set to be created.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_sns.present(name) == ret

        with patch.dict(boto_sns.__opts__, {"test": False}):
            comt = "Failed to create {} AWS SNS topic".format(name)
            ret.update({"comment": comt, "result": False})
            assert boto_sns.present(name) == ret


def test_absent():
    """
    Test to ensure the named sns topic is deleted.
    """
    name = "test.example.com."

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    exists_mock = MagicMock(side_effect=[False, True, True, True, True, True, True])
    with patch.dict(boto_sns.__salt__, {"boto_sns.exists": exists_mock}):
        # tests topic already absent
        comt = "AWS SNS topic {} does not exist.".format(name)
        ret.update({"comment": comt})
        assert boto_sns.absent(name) == ret

        with patch.dict(boto_sns.__opts__, {"test": True}):
            # tests topic present, test option, unsubscribe is False
            comt = (
                "AWS SNS topic {} is set to be removed.  "
                "0 subscription(s) will be removed.".format(name)
            )
            ret.update({"comment": comt, "result": None})
            assert boto_sns.absent(name) == ret

        subscriptions = [
            dict(
                Endpoint="arn:aws:lambda:us-west-2:123456789:function:test",
                Owner=123456789,
                Protocol="Lambda",
                TopicArn="arn:aws:sns:us-west-2:123456789:test",
                SubscriptionArn="arn:aws:sns:us-west-2:123456789:test:some_uuid",
            )
        ]
        with patch.dict(boto_sns.__opts__, {"test": True}):
            subs_mock = MagicMock(return_value=subscriptions)
            with patch.dict(
                boto_sns.__salt__,
                {"boto_sns.get_all_subscriptions_by_topic": subs_mock},
            ):
                # tests topic present, 1 subscription, test option, unsubscribe is True
                comt = (
                    "AWS SNS topic {} is set to be removed.  "
                    "1 subscription(s) will be removed.".format(name)
                )
                ret.update({"comment": comt, "result": None})
                assert boto_sns.absent(name, unsubscribe=True) == ret

        subs_mock = MagicMock(return_value=subscriptions)
        unsubscribe_mock = MagicMock(side_effect=[True, False])
        with patch.dict(boto_sns.__salt__, {"boto_sns.unsubscribe": unsubscribe_mock}):
            with patch.dict(
                boto_sns.__salt__,
                {"boto_sns.get_all_subscriptions_by_topic": subs_mock},
            ):
                delete_mock = MagicMock(side_effect=[True, True, True, False])
                with patch.dict(boto_sns.__salt__, {"boto_sns.delete": delete_mock}):
                    # tests topic present, unsubscribe flag True, unsubscribe succeeded,
                    # delete succeeded
                    comt = "AWS SNS topic {} deleted.".format(name)
                    ret.update(
                        {
                            "changes": {
                                "new": None,
                                "old": {"topic": name, "subscriptions": subscriptions},
                            },
                            "result": True,
                            "comment": comt,
                        }
                    )
                    assert boto_sns.absent(name, unsubscribe=True) == ret

                    # tests topic present, unsubscribe flag True, unsubscribe fails,
                    # delete succeeded
                    ret.update(
                        {
                            "changes": {
                                "new": {"subscriptions": subscriptions},
                                "old": {"topic": name, "subscriptions": subscriptions},
                            },
                            "result": True,
                            "comment": comt,
                        }
                    )
                    assert boto_sns.absent(name, unsubscribe=True) == ret

                    # tests topic present, unsubscribe flag False, delete succeeded
                    ret.update(
                        {
                            "changes": {"new": None, "old": {"topic": name}},
                            "result": True,
                            "comment": comt,
                        }
                    )
                    assert boto_sns.absent(name) == ret

                    # tests topic present, unsubscribe flag False, delete failed
                    comt = "Failed to delete {} AWS SNS topic.".format(name)
                    ret.update({"changes": {}, "result": False, "comment": comt})
                    assert boto_sns.absent(name) == ret



"""
    Test cases for salt.modules.helm
"""


import pytest

import salt.modules.helm as helm

# Import Exception
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {helm: {}}


def test__prepare_cmd():
    assert helm._prepare_cmd() == ("helm",)


def test__prepare_cmd_binary():
    assert helm._prepare_cmd(binary="binary") == ("binary",)


def test__prepare_cmd_commands():
    assert helm._prepare_cmd(commands=["com1", "com2"]) == (
        "helm",
        "com1",
        "com2",
    )


def test__prepare_cmd_flags():
    assert helm._prepare_cmd(flags=["flag1", "--flag2"]) == (
        "helm",
        "--flag1",
        "--flag2",
    )


def test__prepare_cmd_kvflags():
    result_tuple = helm._prepare_cmd(kvflags={"kflag1": "vflag1", "--kflag2": "vflag2"})
    tuple_valide_1 = (
        "helm",
        "--kflag1",
        "vflag1",
        "--kflag2",
        "vflag2",
    )
    tuple_valide_2 = (
        "helm",
        "--kflag2",
        "vflag2",
        "--kflag1",
        "vflag1",
    )

    assert result_tuple == tuple_valide_1 or result_tuple == tuple_valide_2


def test__exec_cmd():
    cmd_prepare = helm._prepare_cmd()
    cmd_prepare_str = " ".join(cmd_prepare)
    cmd_return = {
        "stdout": "succes",
        "stderr": "",
        "retcode": 0,
    }
    result = cmd_return
    result.update({"cmd": cmd_prepare_str})
    with patch.dict(
        helm.__salt__,
        {  # pylint: disable=no-member
            "cmd.run_all": MagicMock(return_value=cmd_return)
        },
    ):
        assert helm._exec_cmd() == result


def test__exec_true_return_valid():
    _exec_cmd_return = {"retcode": 0}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert True is helm._exec_true_return()


def test__exec_true_return_not_valid():
    _exec_cmd_return = {"retcode": -1, "stderr": "test"}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert "test" == helm._exec_true_return()


def test__exec_string_return_valid():
    _exec_cmd_return = {"retcode": 0, "stdout": "test"}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert "test" == helm._exec_string_return()


def test__exec_string_return_not_valid():
    _exec_cmd_return = {"retcode": -1, "stderr": "test"}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert "test" == helm._exec_string_return()


def test__exec_dict_return_valide():
    _exec_cmd_return = {"retcode": 0, "stdout": '{"test": true}'}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert {"test": True} == helm._exec_dict_return()


def test__exec_dict_return_valide_no_json():
    _exec_cmd_return = {"retcode": 0, "stdout": '{"test": true}'}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert '{"test": true}' == helm._exec_dict_return(kvflags={"output": "table"})


def test__exec_dict_return_not_valid():
    _exec_cmd_return = {"retcode": -1, "stderr": "test"}
    with patch("salt.modules.helm._exec_cmd", MagicMock(return_value=_exec_cmd_return)):
        assert "test" == helm._exec_dict_return()


def test_completion():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.completion("bash")
        assert [
            call(commands=["completion", "bash"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_create():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.create("name")
        assert [
            call(commands=["create", "name"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_dependency_build():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.dependency_build("chart")
        assert [
            call(
                commands=["dependency", "build", "chart"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_dependency_list():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.dependency_list("chart")
        assert [
            call(
                commands=["dependency", "list", "chart"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_dependency_update():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.dependency_update("chart")
        assert [
            call(
                commands=["dependency", "update", "chart"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_env():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.env()
        assert [
            call(commands=["env"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_get_all():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.get_all("release")
        assert [
            call(commands=["get", "all", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_get_hooks():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.get_hooks("release")
        assert [
            call(commands=["get", "hooks", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_get_manifest():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.get_manifest("release")
        assert [
            call(
                commands=["get", "manifest", "release"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_get_notes():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.get_notes("release")
        assert [
            call(commands=["get", "notes", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_get_values():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.get_values("release")
        assert [
            call(commands=["get", "values", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_help_():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.help_("command")
        assert [
            call(commands=["help", "command"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_history():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.history("release")
        assert [
            call(commands=["history", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_install():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.install("release", "chart")
        assert [
            call(
                commands=["install", "release", "chart"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_lint():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.lint("path")
        assert [
            call(commands=["lint", "path"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_list_():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.list_()
        assert [
            call(commands=["list"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_package():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.package("chart")
        assert [
            call(commands=["package", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_plugin_install():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.plugin_install("path")
        assert [
            call(commands=["plugin", "install", "path"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_plugin_list():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.plugin_list()
        assert [
            call(commands=["plugin", "list"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_plugin_uninstall():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.plugin_uninstall("plugin")
        assert [
            call(
                commands=["plugin", "uninstall", "plugin"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_plugin_update():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.plugin_update("plugin")
        assert [
            call(
                commands=["plugin", "update", "plugin"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_pull():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.pull("pkg")
        assert [
            call(commands=["pull", "pkg"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_repo_add():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.repo_add("name", "url")
        assert [
            call(
                commands=["repo", "add", "name", "url"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_repo_index():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.repo_index("directory")
        assert [
            call(
                commands=["repo", "index", "directory"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_repo_list():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.repo_list()
        assert [
            call(commands=["repo", "list"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_repo_remove():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.repo_remove("name")
        assert [
            call(commands=["repo", "remove", "name"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_repo_update():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.repo_update()
        assert [
            call(commands=["repo", "update"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_repo_manage_present_bad_list():
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with pytest.raises(CommandExecutionError):
            helm.repo_manage(present=["test"])


def test_repo_manage_present_bad_format():
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with pytest.raises(CommandExecutionError):
            helm.repo_manage(present=[{"test": True}])


def test_repo_manage_present_failed():
    result_wanted = {
        "present": [],
        "added": [],
        "absent": [],
        "removed": [],
        "failed": [{"name": "myname", "url": "myurl"}],
    }
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with patch("salt.modules.helm.repo_add", MagicMock(return_value="failed")):
            assert (
                helm.repo_manage(present=[{"name": "myname", "url": "myurl"}])
                == result_wanted
            )


def test_repo_manage_present_succeed():
    result_wanted = {
        "present": [],
        "added": [{"name": "myname", "url": "myurl"}],
        "absent": [],
        "removed": [],
        "failed": [],
    }
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with patch("salt.modules.helm.repo_add", MagicMock(return_value=True)):
            assert (
                helm.repo_manage(present=[{"name": "myname", "url": "myurl"}])
                == result_wanted
            )


def test_repo_manage_present_already_present():
    result_wanted = {
        "present": [{"name": "myname", "url": "myurl"}],
        "added": [],
        "absent": [],
        "removed": [],
        "failed": [],
    }
    with patch(
        "salt.modules.helm.repo_list",
        MagicMock(return_value=[{"name": "myname", "url": "myurl"}]),
    ):
        assert (
            helm.repo_manage(present=[{"name": "myname", "url": "myurl"}])
            == result_wanted
        )


def test_repo_manage_prune():
    result_wanted = {
        "present": [],
        "added": [],
        "absent": [],
        "removed": ["myname"],
        "failed": [],
    }
    with patch(
        "salt.modules.helm.repo_list",
        MagicMock(return_value=[{"name": "myname", "url": "myurl"}]),
    ):
        with patch("salt.modules.helm.repo_remove", MagicMock(return_value=True)):
            assert helm.repo_manage(prune=True) == result_wanted


def test_repo_manage_absent():
    result_wanted = {
        "present": [],
        "added": [],
        "absent": ["myname"],
        "removed": [],
        "failed": [],
    }
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with patch("salt.modules.helm.repo_remove", MagicMock(return_value=False)):
            assert helm.repo_manage(absent=["myname"]) == result_wanted


def test_repo_manage_removed():
    result_wanted = {
        "present": [],
        "added": [],
        "absent": [],
        "removed": ["myname"],
        "failed": [],
    }
    with patch("salt.modules.helm.repo_list", MagicMock(return_value=None)):
        with patch("salt.modules.helm.repo_remove", MagicMock(return_value=True)):
            assert helm.repo_manage(absent=["myname"]) == result_wanted


def test_rollback():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.rollback("release", "revision")
        assert [
            call(
                commands=["rollback", "release", "revision"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_search_hub():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.search_hub("keyword")
        assert [
            call(commands=["search", "hub", "keyword"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_search_repo():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.search_repo("keyword")
        assert [
            call(commands=["search", "repo", "keyword"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_show_all():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.show_all("chart")
        assert [
            call(commands=["show", "all", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_show_chart():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.show_chart("chart")
        assert [
            call(commands=["show", "chart", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_show_readme():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.show_readme("chart")
        assert [
            call(commands=["show", "readme", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_show_values():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.show_values("chart")
        assert [
            call(commands=["show", "values", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_status():
    magic_mock = MagicMock(return_value={"test": True})
    with patch("salt.modules.helm._exec_dict_return", magic_mock):
        assert {"test": True} == helm.status("release")
        assert [
            call(commands=["status", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_template():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.template("name", "chart")
        assert [
            call(commands=["template", "name", "chart"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_test():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.test("release")
        assert [
            call(commands=["test", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_uninstall():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.uninstall("release")
        assert [
            call(commands=["uninstall", "release"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_upgrade():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.upgrade("release", "chart")
        assert [
            call(
                commands=["upgrade", "release", "chart"],
                flags=None,
                kvflags=None,
            )
        ] == magic_mock.mock_calls


def test_verify():
    magic_mock = MagicMock(return_value=True)
    with patch("salt.modules.helm._exec_true_return", magic_mock):
        assert True is helm.verify("path")
        assert [
            call(commands=["verify", "path"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls


def test_version():
    magic_mock = MagicMock(return_value="the_return")
    with patch("salt.modules.helm._exec_string_return", magic_mock):
        assert "the_return" == helm.version()
        assert [
            call(commands=["version"], flags=None, kvflags=None)
        ] == magic_mock.mock_calls



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest

import salt.modules.boto_route53 as boto53mod
import salt.modules.boto_vpc as botovpcmod
import salt.states.boto_route53 as boto_route53
from tests.support.mock import MagicMock, create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_route53: {}}


@pytest.fixture
def patch_botomod_hosted_zones():
    with patch.dict(
        boto_route53.__salt__,
        {
            "boto_route53.describe_hosted_zones": create_autospec(
                boto53mod.describe_hosted_zones
            ),
            "boto_vpc.describe_vpcs": create_autospec(botovpcmod.describe_vpcs),
        },
    ):
        yield


@pytest.fixture
def fake_single_vpc(patch_botomod_hosted_zones):
    boto_route53.__salt__["boto_vpc.describe_vpcs"].return_value = {
        "vpcs": [{"region": "fnordland", "id": "fnord"}],
    }
    boto_route53.__salt__["boto_route53.describe_hosted_zones"].return_value = {
        "HostedZone": {"Config": {"PrivateZone": "true"}},
        "VPCs": {"VPC": {"VPCId": "fnord", "VPCRegion": "fnordland"}},
    }


@pytest.fixture
def fake_multiple_vpcs(patch_botomod_hosted_zones):
    boto_route53.__salt__["boto_vpc.describe_vpcs"].return_value = {
        "vpcs": [{"region": "fnordland", "id": "fnord"}],
    }
    boto_route53.__salt__["boto_route53.describe_hosted_zones"].return_value = {
        "HostedZone": {"Config": {"PrivateZone": "true"}},
        "VPCs": [
            {"VPCId": "fnord", "VPCRegion": "fnordland"},
            {"VPCId": "fnord part 2", "VPCRegion": "fnordlandia"},
        ],
    }


def test_present():
    """
    Test to ensure the Route53 record is present.
    """
    name = "test.example.com."
    value = "1.1.1.1"
    zone = "example.com."
    record_type = "A"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[{}, {}, {"value": ""}, False])
    mock_bool = MagicMock(return_value=False)
    with patch.dict(
        boto_route53.__salt__,
        {"boto_route53.get_record": mock, "boto_route53.add_record": mock_bool},
    ):
        with patch.dict(boto_route53.__opts__, {"test": False}):
            comt = "Failed to add {} Route53 record.".format(name)
            ret.update({"comment": comt})
            assert boto_route53.present(name, value, zone, record_type) == ret

        with patch.dict(boto_route53.__opts__, {"test": True}):
            comt = "Route53 record {} set to be added.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_route53.present(name, value, zone, record_type) == ret

            comt = "Route53 record {} set to be updated.".format(name)
            ret.update({"comment": comt})
            assert boto_route53.present(name, value, zone, record_type) == ret

        ret.update({"comment": "", "result": True})
        assert boto_route53.present(name, value, zone, record_type) == ret


def test_absent():
    """
    Test to ensure the Route53 record is deleted.
    """
    name = "test.example.com."
    zone = "example.com."
    record_type = "A"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(boto_route53.__salt__, {"boto_route53.get_record": mock}):
        comt = "{} does not exist.".format(name)
        ret.update({"comment": comt})
        assert boto_route53.absent(name, zone, record_type) == ret

        with patch.dict(boto_route53.__opts__, {"test": True}):
            comt = "Route53 record {} set to be deleted.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_route53.absent(name, zone, record_type) == ret


def test_hosted_zone_present_should_not_fail_when_one_vpc_in_deets(fake_single_vpc):
    boto_route53.hosted_zone_present(
        name="fnord", private_zone=True, vpc_region="fnordland", vpc_name="fnord"
    )


def test_hosted_zone_present_should_not_fail_with_multiple_vpcs_in_deets(
    fake_multiple_vpcs,
):
    boto_route53.hosted_zone_present(
        name="fnord", private_zone=True, vpc_region="fnordland", vpc_name="fnord"
    )



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""

import copy
import logging

import pytest

import salt.utils.platform
from salt import client
from salt.exceptions import (
    EauthAuthenticationError,
    SaltClientError,
    SaltInvocationError,
    SaltReqTimeoutError,
)
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


def test_job_result_return_success(master_opts):
    """
    Should return the `expected_return`, since there is a job with the right jid.
    """
    minions = ()
    jid = "0815"
    raw_return = {"id": "fake-id", "jid": jid, "data": "", "return": "fake-return"}
    expected_return = {"fake-id": {"ret": "fake-return"}}
    with client.LocalClient(mopts=master_opts) as local_client:
        local_client.event.get_event = MagicMock(return_value=raw_return)
        local_client.returners = MagicMock()
        ret = local_client.get_event_iter_returns(jid, minions)
        val = next(ret)
        assert val == expected_return


def test_job_result_return_failure(master_opts):
    """
    We are _not_ getting a job return, because the jid is different. Instead we should
    get a StopIteration exception.
    """
    minions = ()
    jid = "0815"
    raw_return = {
        "id": "fake-id",
        "jid": "0816",
        "data": "",
        "return": "fake-return",
    }
    with client.LocalClient(mopts=master_opts) as local_client:
        local_client.event.get_event = MagicMock()
        local_client.event.get_event.side_effect = [raw_return, None]
        local_client.returners = MagicMock()
        ret = local_client.get_event_iter_returns(jid, minions)
        with pytest.raises(StopIteration):
            next(ret)


def test_create_local_client(master_opts):
    with client.LocalClient(mopts=master_opts) as local_client:
        assert isinstance(
            local_client, client.LocalClient
        ), "LocalClient did not create a LocalClient instance"


def test_check_pub_data(salt_master_factory):
    just_minions = {"minions": ["m1", "m2"]}
    jid_no_minions = {"jid": "1234", "minions": []}
    valid_pub_data = {"minions": ["m1", "m2"], "jid": "1234"}

    config = copy.deepcopy(salt_master_factory.config)
    salt_local_client = salt.client.get_local_client(mopts=config)

    pytest.raises(EauthAuthenticationError, salt_local_client._check_pub_data, "")
    assert {} == salt_local_client._check_pub_data(
        just_minions
    ), "Did not handle lack of jid correctly"

    assert {} == salt_local_client._check_pub_data(
        {"jid": "0"}
    ), "Passing JID of zero is not handled gracefully"

    with patch.dict(salt_local_client.opts, {}):
        salt_local_client._check_pub_data(jid_no_minions)

    assert valid_pub_data == salt_local_client._check_pub_data(valid_pub_data)


def test_cmd_subset(salt_master_factory):
    salt_local_client = salt.client.get_local_client(mopts=salt_master_factory.config)

    with patch(
        "salt.client.LocalClient.cmd",
        return_value={
            "minion1": ["first.func", "second.func"],
            "minion2": ["first.func", "second.func"],
        },
    ):
        with patch("salt.client.LocalClient.cmd_cli") as cmd_cli_mock:
            salt_local_client.cmd_subset("*", "first.func", subset=1, cli=True)
            try:
                cmd_cli_mock.assert_called_with(
                    ["minion2"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=False,
                    ret="",
                )
            except AssertionError:
                cmd_cli_mock.assert_called_with(
                    ["minion1"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=False,
                    ret="",
                )
            salt_local_client.cmd_subset("*", "first.func", subset=10, cli=True)
            try:
                cmd_cli_mock.assert_called_with(
                    ["minion2", "minion1"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=False,
                    ret="",
                )
            except AssertionError:
                cmd_cli_mock.assert_called_with(
                    ["minion1", "minion2"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=False,
                    ret="",
                )

            ret = salt_local_client.cmd_subset(
                "*", "first.func", subset=1, cli=True, full_return=True
            )
            try:
                cmd_cli_mock.assert_called_with(
                    ["minion2"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=True,
                    ret="",
                )
            except AssertionError:
                cmd_cli_mock.assert_called_with(
                    ["minion1"],
                    "first.func",
                    (),
                    progress=False,
                    kwarg=None,
                    tgt_type="list",
                    full_return=True,
                    ret="",
                )


@pytest.mark.skip_on_windows(reason="Not supported on Windows")
def test_pub(salt_master_factory):
    """
    Tests that the client cleanly returns when the publisher is not running

    Note: Requires ZeroMQ's IPC transport which is not supported on windows.
    """
    config = copy.deepcopy(salt_master_factory.config)
    salt_local_client = salt.client.get_local_client(mopts=config)

    if salt_local_client.opts.get("transport") != "zeromq":
        pytest.skip("This test only works with ZeroMQ")
    # Make sure we cleanly return if the publisher isn't running
    with patch("os.path.exists", return_value=False):
        pytest.raises(SaltClientError, lambda: salt_local_client.pub("*", "test.ping"))

    # Check nodegroups behavior
    with patch("os.path.exists", return_value=True):
        with patch.dict(
            salt_local_client.opts,
            {
                "nodegroups": {
                    "group1": "L@foo.domain.com,bar.domain.com,baz.domain.com or bl*.domain.com"
                }
            },
        ):
            # Do we raise an exception if the nodegroup can't be matched?
            pytest.raises(
                SaltInvocationError,
                salt_local_client.pub,
                "non_existent_group",
                "test.ping",
                tgt_type="nodegroup",
            )


@pytest.mark.skip_unless_on_windows(reason="Windows only test")
@pytest.mark.slow_test
def test_pub_win32(salt_master_factory):
    """
    Tests that the client raises a timeout error when using ZeroMQ's TCP
    transport and publisher is not running.

    Note: Requires ZeroMQ's TCP transport, this is only the default on Windows.
    """
    config = copy.deepcopy(salt_master_factory.config)
    salt_local_client = salt.client.get_local_client(mopts=config)

    if salt_local_client.opts.get("transport") != "zeromq":
        pytest.skip("This test only works with ZeroMQ")
    # Make sure we cleanly return if the publisher isn't running
    with patch("os.path.exists", return_value=False):
        pytest.raises(
            SaltReqTimeoutError, lambda: salt_local_client.pub("*", "test.ping")
        )

    # Check nodegroups behavior
    with patch("os.path.exists", return_value=True):
        with patch.dict(
            salt_local_client.opts,
            {
                "nodegroups": {
                    "group1": "L@foo.domain.com,bar.domain.com,baz.domain.com or bl*.domain.com"
                }
            },
        ):
            # Do we raise an exception if the nodegroup can't be matched?
            pytest.raises(
                SaltInvocationError,
                salt_local_client.pub,
                "non_existent_group",
                "test.ping",
                tgt_type="nodegroup",
            )
